import numpy as np
import bionumpy as bnp

def bionumpy_hash(numeric_sequence, k):
    # reversing stuff to be compatible with old hashing
    #kmers = fast_hash(numeric_sequence.astype(np.uint8)[::-1], k)[::-1]
    numeric_sequence = numeric_sequence   # [::-1]
    encoded_sequence = bnp.EncodedArray(numeric_sequence, bnp.encodings.alphabet_encoding.ACTGEncoding)
    kmers = bnp.sequence.get_kmers(encoded_sequence, k).raw()   # [::-1]  # raw to get numpy array
    return kmers








import logging
import numpy as np


class CountMinSketchKmerCounter:
    def __init__(self, data, modulos):
        self._data = data
        self._modulos = modulos
        self._array_positions = np.concatenate([[0], np.cumsum(modulos)[:-1]])

    def _indexes(self, kmer):
        return (kmer % self._modulos) + self._array_positions

    def get_count(self, kmer):
        return np.min(self._data[self._indexes(kmer)])

    def get_counts(self, kmers):
        pass

    @classmethod
    def create_empty(cls, modulos, dtype=np.uint16):
        data = np.zeros(np.sum(modulos), dtype=dtype)
        return cls(data, np.asanyarray(modulos))

    def count_kmers(self, kmers):
        for mod, offset in zip(self._modulos, self._array_positions):
            indexes = (kmers % mod) + offset
            counts = np.bincount(indexes).astype(self._data.dtype)
            self._data[0:len(counts)] += counts



import logging
import time
from .kmer_hashing import power_array

logging.basicConfig(level=logging.INFO)
import numpy as np
from .flat_kmers import FlatKmers, letter_sequence_to_numeric
from Bio.Seq import Seq
from collections import defaultdict
from .kmer_hashing import reverse_power_array, power_array, kmer_hashes_to_bases
from .flat_kmers import numeric_to_letter_sequence


def kmer_hash_to_sequence(hash, k):
    bases = kmer_hashes_to_bases(np.array([hash]), k)[0]
    return ''.join([b for b in numeric_to_letter_sequence(bases)])


def sequence_to_kmer_hash(sequence):
    return kmer_to_hash_fast(letter_sequence_to_numeric(sequence).astype(np.uint64), len(sequence))


#@jit(nopython=True)
def kmer_to_hash_fast(kmer, k):
    assert kmer.dtype == np.uint64
    return int(np.sum(kmer * reverse_power_array(k)))


class SnpKmerFinder:
    """
    Simple kmer finder that only supports SNP graphs
    """

    def __init__(self, graph, k=15, spacing=None, include_reverse_complements=False, pruning=False, max_kmers_same_position=100000,
                 max_frequency=10000, max_variant_nodes=10000, only_add_variant_kmers=False, whitelist=None, only_save_variant_nodes=False,
                 start_position=None, end_position=None, only_store_nodes=None, skip_kmers_with_nodes=None, only_save_one_node_per_kmer=False,
                 reference=None, variant_to_nodes=None, node_to_variants=None, haplotype_matrix=None):
        self.graph = graph
        self.reference = reference
        self.k = k
        #logging.info("Getting linear ref nodes")
        #self.linear_nodes = graph.linear_ref_nodes()


        self._hashes = []
        self._nodes = []
        self._ref_offsets = []
        self._allele_frequencies = []
        self.kmers_found = []
        self._bases_in_search_path = []
        self._nodes_in_path = []
        self._kmers_found = 0
        self._current_ref_offset = None
        self._last_ref_pos_added = 0
        self.pruning = pruning
        self._n_kmers_skipped_low_allele_frequency = 0
        self._n_kmers_pruned = 0
        self._has_traversed_variant = False
        self._unique_kmers_added = set()
        self._max_kmers_same_position = max_kmers_same_position
        self._n_kmers_added_current_position = 0
        self._n_kmers_skipped = 0
        self._kmer_frequencies = defaultdict(int)
        self._max_frequency = max_frequency
        self._n_skipped_due_to_frequency = 0
        self._max_variant_nodes = max_variant_nodes
        self._n_skipped_due_to_max_variant_nodes = 0
        self._only_add_variant_kmers = only_add_variant_kmers
        self._whitelist = whitelist
        self._n_skipped_whitelist = 0
        self._start_position = start_position
        self._end_position = end_position
        self._only_store_nodes = only_store_nodes
        self._skip_kmers_with_nodes = skip_kmers_with_nodes
        self._n_skipped_blacklist_nodes = 0
        self._only_save_one_node_per_kmer = only_save_one_node_per_kmer
        self.haplotype_matrix = haplotype_matrix
        self.variant_to_nodes = variant_to_nodes
        self.node_to_variants = node_to_variants

        if self._start_position == None:
            self._start_position = 0

        self._only_save_variant_nodes = only_save_variant_nodes
        self._variant_nodes = set()
        if self._only_save_variant_nodes:
            logging.info("Will only store nodes associated to variants for each kmer")
            # Find all variant nodes, store in a set
            for node in range(len(graph.nodes)):
                if node % 100000 == 0:
                    logging.info("Adding variant nodes, on node %d" % node)
                if len(graph.get_edges(node)) > 1:
                    for next_node in graph.get_edges(node):
                        self._variant_nodes.add(next_node)

        if spacing is None:
            self.spacing = k
        else:
            self.spacing = spacing
        self._include_reverse_complements=include_reverse_complements
        if self._include_reverse_complements:
            logging.info("Will include reverse complement of kmers")


    def has_kmer(self, kmer, nodes):
        for found in self.kmers_found:
            if found[0] == kmer and found[1] == nodes:
                return True
        return False


        if (kmer, nodes) in self.kmers_found:
            return True
        return False


    def _add_kmer(self, kmer, nodes):
        #logging.info("      Addign kmer %s with nodes %s" % (kmer, nodes))
        self._n_paths_searched += 1

        assert len(kmer) == self.k

        hash = kmer_to_hash_fast(letter_sequence_to_numeric(kmer), k=len(kmer))

        if self._whitelist is not None:
            # Either hash or reverse complement hash can be in whitelist in order to keep this kmer
            if hash not in self._whitelist:
                rev_hash = kmer_to_hash_fast(letter_sequence_to_numeric(str(Seq(kmer).reverse_complement())),
                                             k=len(kmer))
                if rev_hash not in self._whitelist:
                    self._n_skipped_whitelist += 1
                    return

        if self._skip_kmers_with_nodes is not None and len(set(nodes).intersection(self._skip_kmers_with_nodes)) > 0:
            self._n_skipped_blacklist_nodes += 1
            return

        if not self._has_traversed_variant and self._only_add_variant_kmers:
            return

        if self._kmer_frequencies[hash] >= self._max_frequency:
            self._n_skipped_due_to_frequency += 1
            return

        if self.pruning and hash not in self._unique_kmers_added:
            if self._last_ref_pos_added != self._current_ref_offset and self._last_ref_pos_added > self._current_ref_offset - 124:
                if not self._has_traversed_variant:
                    # Do not add
                    self._n_kmers_pruned += 1
                    return

        if self._n_kmers_added_current_position > self._max_kmers_same_position:
            self._n_kmers_skipped += 1
            return

        n_variant_nodes = len([n for n in nodes if not self.graph.is_linear_ref_node_or_linear_ref_dummy_node(n)])
        if n_variant_nodes >= self._max_variant_nodes:
            self._n_skipped_due_to_max_variant_nodes += 1
            logging.info("SKIPPING KMER DUE TO TOO MANY VARIANT NODES")
            return

        if self._include_reverse_complements:
            rev_hash = kmer_to_hash_fast(letter_sequence_to_numeric(str(Seq(kmer).reverse_complement())), k=len(kmer))

        self._unique_kmers_added.add(hash)
        self._kmer_frequencies[hash] += 1


        if self.haplotype_matrix is not None:
            kmer_allele_frequency = self.haplotype_matrix.get_allele_frequency_for_nodes(nodes, self.node_to_variants, self.variant_to_nodes)
        else:
            kmer_allele_frequency = min([self.graph.get_node_allele_frequency(node) for node in nodes])

        for node in nodes:
            if self._only_save_variant_nodes and node not in self._variant_nodes:
                continue

            if self._only_store_nodes is not None and node not in self._only_store_nodes:
                continue

            self._hashes.append(hash)
            self._nodes.append(node)
            self._ref_offsets.append(self._current_ref_offset)
            self._allele_frequencies.append(kmer_allele_frequency)

            if self._include_reverse_complements:
                self._hashes.append(rev_hash)
                self._nodes.append(node)
                self._ref_offsets.append(self._current_ref_offset)
                self._allele_frequencies.append(kmer_allele_frequency)

            if self._only_save_one_node_per_kmer:
                break

        self._last_ref_pos_added = self._current_ref_offset

        self._kmers_found += 1
        if len(self.kmers_found) < 500:
            # Only add to this when there is little data, only used for testing and debugging
            #loggign.info("------Added kmer with nodes %s" % nodes)
            self.kmers_found.append((kmer, nodes, self._current_ref_offset, hash))

        self._n_kmers_added_current_position += 1

    def _find_all_variant_kmers_from_position(self, linear_ref_pos):
        self._n_paths_searched = 0
        self._n_variant_nodes_passed = 0
        self._current_ref_offset = linear_ref_pos

        # Always start one base pair before, but do not include that base pair
        # this lets us catch cases where we are at a beginning of a node

        if linear_ref_pos > 0:
            node = self.graph.get_node_at_ref_offset(linear_ref_pos-1)
            offset = self.graph.get_node_offset_at_ref_offset(linear_ref_pos-1) + 1
        else:
            node = self.graph.get_node_at_ref_offset(linear_ref_pos)
            offset = self.graph.get_node_offset_at_ref_offset(linear_ref_pos)

        self._bases_in_search_path = []
        self._nodes_in_path = []
        self._has_traversed_variant = False
        self._n_kmers_added_current_position = 0
        self._search_graph_from(node, offset, self.k)



    def _search_graph_from(self, node, offset, bases_left):
        #loggign.info("== Searching node %d from offset %d. Bases left: %d == " % (node, offset, bases_left))

        if False and self._n_paths_searched > 1000:
            logging.warning("More than 1000 paths searched from ref pos %d. Possibly many variants here?" % self._current_ref_offset)

        if bases_left == 0:
            self._add_kmer(''.join(self._bases_in_search_path).replace("-", ""), set(self._nodes_in_path))
            #loggign.info("Recursion is done, added kmer with nodes: %s" % str(self._nodes_in_path))
            return

        if not self.graph.is_linear_ref_node_or_linear_ref_dummy_node(node):
            self._n_variant_nodes_passed += 1

        # Process the rest of this node
        node_size = int(self.graph.nodes[node])
        #node_sequence = self.graph.get_node_sequence(node)
        node_sequence = self.graph.get_node_subsequence(node, offset, min(offset + bases_left, node_size))
        #loggign.info("Got node sequence for node %d, from offset %d to %d: %s" % (node, offset, min(offset + bases_left, node_size), node_sequence))

        # Special case for empty node (without sequence, we always want to add this
        if node_size == 0:
            # add a dummy sequence
            self._bases_in_search_path.append("-")
            self._nodes_in_path.append(node)


        for node_position in range(int(offset), node_size):
            base = node_sequence[node_position-int(offset)]  # subtract offset since node sequence now starts at offset
            #loggign.info("   Adding base %s at node %d, pos %d" % (base, node, node_position))
            self._bases_in_search_path.append(base)
            self._nodes_in_path.append(node)
            bases_left -= 1

            if bases_left == 0:
                # Remove dummy sequence "-" when adding kmer. These are added on indel nodes
                self._add_kmer(''.join(self._bases_in_search_path).replace("-", ""), set(self._nodes_in_path))
                #loggign.info("Recursion is done, added kmer with nodes: %s" % str(self._nodes_in_path))
                # Recursion is done
                return

        # If offset is last base in this node, recursively search next nodes depth first
        next_nodes = self.graph.get_edges(node)
        if len(next_nodes) > 1:
            self._has_traversed_variant = True

        # Prioritize linear ref node
        if len(next_nodes) > 0 and not self.graph.is_linear_ref_node_or_linear_ref_dummy_node(next_nodes[0]):
            next_nodes = list(reversed(next_nodes))

        if len(next_nodes) > 1:
            if self._n_variant_nodes_passed >= self._max_variant_nodes:
                #loggign.info("Passet too many variant nodes, only following linear ref at next nodes %s" % (next_nodes))
                # Only choose linear ref node
                next_nodes = [n for n in next_nodes if self.graph.is_linear_ref_node_or_linear_ref_dummy_node(n)]

        bases_so_far = len(self._bases_in_search_path)
        #loggign.info("Next nodes: %s" % str(next_nodes))
        for next_node in next_nodes:
            self._search_graph_from(next_node, 0, bases_left)
            self._bases_in_search_path = self._bases_in_search_path[0:bases_so_far]
            self._nodes_in_path = self._nodes_in_path[0:bases_so_far]


    def find_kmers_from_linear_ref_position(self, pos):
        self._find_all_variant_kmers_from_position(pos)

    def get_flat_kmers(self, v=None):
        return FlatKmers(np.array(self._hashes, dtype=np.uint64), np.array(self._nodes, np.uint32), np.array(self._ref_offsets, np.uint64), np.array(self._allele_frequencies, np.single))

    def find_kmers_on_linear_reference(self):
        #assert self.spacing == 1, "Finding kmers on linear reference is only possible when spacing is 1"
        #loggign.info("Fetching reference sequence between %d and %d" % (self._start_position, self._end_position+self.k))
        reference_sequence = str(self.reference[self._start_position:self._end_position+self.k])
        assert len(reference_sequence) > 0, "No reference sequence between positions %d and %d" % (self._start_position, self._end_position+self.k)
        #loggign.info("Fetching kmers")
        from .read_kmers import ReadKmers
        kmers = ReadKmers.get_kmers_from_read_dynamic(reference_sequence, power_array(self.k))
        kmers = kmers[::self.spacing]
        #loggign.info("Done fetching kmers")

        self._hashes = kmers
        self._nodes = np.zeros(len(kmers)) + 1
        self._ref_offsets = np.arange(self._start_position, self._start_position+len(reference_sequence), self.spacing)
        self._allele_frequencies = np.zeros(len(kmers)) + 1.0

    def find_kmers(self):
        if self.reference is not None:
            logging.warning("Will find kmers on linear reference and not graph")
            self.find_kmers_on_linear_reference()
            return self.get_flat_kmers()

        #loggign.info("Linear reference is %d bp" % self.graph.linear_ref_length())
        if self._end_position is None:
            self._end_position = self.graph.linear_ref_length()


        prev_time = time.time()
        for i in range(self._start_position // self.spacing, self.graph.linear_ref_length() // self.spacing):
            pos = i * self.spacing
            if i % 50000 == 0:

                logging.info("On ref position %d/%s. Time spent: %.3f. %d/%d basepairs traversed. %d kmers found. Have pruned %d kmers. "
                             "Skipped %d kmers. Skipped due to high frequency: %d. Skipped because too many variant nodes: %d. Skipped because blacklisted node: %d. Skipped low allele freq: %d"
                             % (pos, self._end_position, time.time() - prev_time, (i - self._start_position//self.spacing) * self.spacing, self._end_position-self._start_position,
                                self._kmers_found, self._n_kmers_pruned, self._n_kmers_skipped, self._n_skipped_due_to_frequency, self._n_skipped_due_to_max_variant_nodes, self._n_skipped_blacklist_nodes,
                                self._n_kmers_skipped_low_allele_frequency))

                prev_time = time.time()
                if self._whitelist is not None:
                    logging.info("N skipped because not in whitelist: %d" % self._n_skipped_whitelist)
            if self._end_position is not None and pos >= self._end_position:
                logging.info("Ending at end position %d" % self._end_position)
                break
            self.find_kmers_from_linear_ref_position(pos)


        return self.get_flat_kmers()


if __name__ == "__main__":

    test_simple()



import numpy as np


class NpList:
    """
    Simple List-like data structure backed by a numpy array
    """
    def __init__(self, dtype=None):
        self._data = np.empty(0)
        if dtype is not None:
            self._data = self._data.astype(dtype)
        self._dtype = dtype
        self._n_elements = 0

    def _initialize_array(self, first_element):
        if self._dtype is None:
            self._dtype = type(first_element)
            self._data = self._data.astype(self._dtype)
        self._data = np.zeros(100, dtype=self._dtype)

    def _extend_data(self, new_length):
        new_data = np.zeros(new_length, dtype=self._data.dtype)
        new_data[0:len(self._data)] = self._data
        self._data = new_data

    def append(self, element):
        if len(self._data) == 0:
            self._initialize_array(element)

        # extend data array if necessary
        if self._n_elements == len(self._data):
           self._extend_data(len(self._data)*2)

        self._data[self._n_elements] = element
        self._n_elements += 1

    def extend(self, elements):
        if self._data is None:
            self._initialize_array(elements[0])
        if self._n_elements + len(elements) >= len(self._data):
            self._extend_data((self._n_elements+len(elements))*2)
        self._data[self._n_elements:self._n_elements+len(elements)] = elements
        self._n_elements += len(elements)

    def __getitem__(self, item):
        return self.get_nparray()[item]  # works for when item is negative, etc

    def get_nparray(self):
        return self._data[0:self._n_elements]

    def set_n_elements(self, n):
        self._n_elements = n

    def copy(self):
        new = NpList(dtype=self._dtype)
        new.extend(self.get_nparray())
        return new

    def __eq__(self, other):
        return np.all(self.get_nparray() == other.get_nparray())

    def __len__(self):
        return self._n_elements

    def __str__(self):
        return str(self.get_nparray())

    def __repr__(self):
        return "NpList(" + str(self) + ")"



import logging
import numpy as np

def power_array(k):
    return np.power(4, np.arange(k-1, -1, -1)).astype(np.uint64)


def reverse_power_array(k):
    return np.power(4, np.arange(k)).astype(np.uint64)


def kmer_hash_to_reverse_complement_hash(hash, k):
    return kmer_hashes_to_reverse_complement_hash(np.array([hash]), k)[0]


def kmer_hashes_to_reverse_complement_hash(hashes, k):
    assert k <= 31
    complement_bases = kmer_hashes_to_complement_bases(hashes, k)
    reverse_complement_hash = np.sum(complement_bases * power_array(k), axis=1)
    return reverse_complement_hash


def kmer_hashes_to_complement_hashes(hashes, k):
    assert k <= 31
    power_array = reverse_power_array(k)
    complement_bases = kmer_hashes_to_complement_bases(hashes, k)
    complement_hash = np.sum(complement_bases * power_array, axis=1)
    return complement_hash



def kmer_hashes_to_complement_bases(hashes, k):
    #raise NotImplementedError("Must be fixed to work with ACGT encoding")
    bases = kmer_hashes_to_bases(hashes, k)
    complement_bases = np.zeros_like(bases)
    complement_bases[bases == 0] = 3
    complement_bases[bases == 3] = 0
    complement_bases[bases == 2] = 1
    complement_bases[bases == 1] = 2
    #complement_bases = (bases + 2) % 4
    return complement_bases



def kmer_hashes_to_bases(hashes, k):
    hashes = hashes.astype(np.uint64)
    bases = np.zeros((len(hashes), k), dtype=np.uint64)
    #for i in range(k-1, -1, -1):
    for i in range(k):
        # print("Hash now: %d" % hash)
        exponential = np.power(np.uint64(4), np.uint64(k - i - 1), dtype=np.uint64)
        # base = hash // exponential   # gives float, no good
        base = np.floor_divide(hashes, exponential, dtype=np.uint64)
        hashes -= base * exponential

        bases[:, i] = base
    return bases[:,::-1]  # reverse after hashing right-to-left




import logging
import time


from npstructures import Counter, HashTable
import numpy as np


class KmerCounter:
    def __init__(self, counter):
        self.counter = counter

    @classmethod
    def from_flat_kmersv2(cls, flat, modulo, subsample_ratio=1):
        logging.info("Subsampling ration: %d" % subsample_ratio)
        kmers = flat._hashes
        logging.info("Kmers beforesubsampling: %d" % len(kmers))
        kmers = kmers[::subsample_ratio]
        logging.info("Kmers after subsampling: %d" % len(kmers))
        logging.info("Finding unique kmers")
        t = time.perf_counter()
        unique_kmers, counts = np.unique(kmers, return_counts=True)
        logging.info("DOne finding unique kmers")
        counter = HashTable(unique_kmers, counts, mod=modulo)
        return cls(counter)

    @classmethod
    def from_flat_kmers(cls, flat, modulo, chunk_size=50000000):
        kmers = flat._hashes
        logging.info("Finding unique kmers")
        t = time.perf_counter()
        unique_kmers = np.unique(kmers)
        logging.info("Time spent to get unique kmers from %d kmers: %d sec" % (len(kmers), time.perf_counter()-t))
        logging.info("Unique kmers: %d" % len(unique_kmers))

        np.save("debugging", unique_kmers)
        logging.info("Saved debugging file")

        logging.info("Making counter")
        t = time.perf_counter()
        counter = Counter(unique_kmers, mod=modulo)
        logging.info("Time spent making counter: %d" % (time.perf_counter()-t))

        logging.info("Counting")
        t = time.perf_counter()
        for i, chunk in enumerate(np.array_split(kmers, 1 + len(kmers) // chunk_size)):
            logging.info("Counting chunk %d" % i)
            counter.count(chunk)

        #counter.count(kmers)
        logging.info("Time spent counting: %d" % (time.perf_counter()-t))
        return cls(counter)

    def get_frequency(self, kmer):
        #assert isinstance(kmer, int), "Kmer is %s" % type(kmer)
        return self.counter[int(kmer)]




import logging
import numpy as np
from .flat_kmers import FlatKmers2, FlatKmers
from .critical_graph_paths import CriticalGraphPaths
from .nplist import NpList
import sys
sys.setrecursionlimit(20000)
from obgraph.position_id import PositionId
from . import kmer_hash_to_sequence
from .kmer_hashing import power_array, reverse_power_array

_complement_lookup = np.array([3, 2, 1, 0], dtype=np.uint64)


def update_hash(current_base, current_hash, first_base, k, only_add=False):
    # only_add!=True means to build hash without subtracting previous first base
    # only_add should then be the number of the base we are at
    # very important that everything is int (not float) when working with large k
    current_hash = int(current_hash)
    #assert type(current_hash) == int, "Current hash has type %s" % type(current_hash)
    current_base = int(current_base)
    first_base = int(first_base)
    current_base_complement = _complement_lookup[current_base]  # (current_base + 2) % 4
    #print("Current base: %d. Only add: %s" % (current_base, only_add))
    if not isinstance(only_add, bool):
        #current_hash = current_hash * 4 + current_base
        current_hash = current_hash + 4**(only_add) * current_base
        #print("ONLY ADDING HASH. Current hash: %d, only add: %d" % (current_hash, only_add))
    else:
        #current_hash = (current_hash - first_base * 4 ** (k - 1)) * 4 + current_base
        current_hash = (current_hash - first_base) // 4 + current_base * 4**(k-1)
        first_base_complement = _complement_lookup[first_base]  # (first_base + 2) % 4

    return current_hash


class DenseKmerFinder:
    """
    Finds all possible kmers in graph
    """
    def __init__(self, graph, k, critical_graph_paths=None,
                 position_id=None, only_save_one_node_per_kmer=False, max_variant_nodes=4,
                 only_store_variant_nodes=False,
                 start_at_critical_path_number=None, stop_at_critical_path_number=None,
                 whitelist=None,
                 only_store_nodes=None,
                 only_follow_nodes=None):

        self._graph = graph
        self._linear_ref_nodes = self._graph.linear_ref_nodes()
        self._k = k
        self._only_save_one_node_per_kmer = only_save_one_node_per_kmer
        self._max_variant_nodes = max_variant_nodes
        self._start_nodes = NpList(dtype=np.int32)
        self._start_offsets = NpList(dtype=np.int16)
        self._nodes = NpList(dtype=np.int32)
        self._kmers = NpList(dtype=np.int64)
        self._allele_frequencies = NpList(dtype=float)
        
        self._power_vector = power_array(k)

        self._current_search_start_node = self._graph.get_first_node()
        self._current_search_start_offset = 0
        self._current_hash = 0
        self._current_bases = NpList(dtype=np.int8)
        self._current_nodes = NpList(dtype=np.int32)

        self.results = []

        self._n_nodes_skipped_because_too_complex = 0
        self._only_store_nodes = only_store_nodes
        self.kmers_found = []

        self._only_store_variant_nodes = only_store_variant_nodes
        if self._only_store_variant_nodes:
            assert variant_to_nodes is not None

        self._critical_graph_paths = critical_graph_paths

        self._position_id = position_id
        if position_id is None:
            logging.warning("Position id index is not set, creating")
            self._position_id = PositionId.from_graph(self._graph)

        self._nonempty_bases_traversed = 0

        self._effective_k = self._k  # how long we want to look back in path. Will be larger than x for empty nodes with dummy bases
        self._positions_treated = set()
        self._current_path_start_position = 0
        self._recursion_depth = 0
        
        self._stop_at_critical_path_number = stop_at_critical_path_number
        self._start_at_critical_path_number = start_at_critical_path_number

        self._whitelist = whitelist  # set of kmers, do not store any other than these
        self._n_skipped_whitelist = 0
        self._only_follow_nodes = only_follow_nodes

        self._early_stop = False

        self._whitelist = None
        if whitelist is not None:
            self._whitelist = whitelist
            #logging.info("Will limit kmers to whitelist (%d kmers in whitelist)" % len(whitelist))

    def get_found_kmers_and_nodes(self):
        return (self._kmers.get_nparray(), self._nodes.get_nparray())

    def get_flat_kmers(self, v="2"):
        assert self._allele_frequencies.get_nparray().dtype == float
        if v == "0" or v == "1":
            #logging.info("Converting start nodes/offsets to an iD to be compatible with FlatKmers")
            # return old version, convert start nodes and offsets to a position id
            start_nodes = self._start_nodes.get_nparray()
            start_offsets = self._start_offsets.get_nparray()
            if v == "1":
                ref_offsets = self._position_id.get(start_nodes, start_offsets)
            else:
                ref_offsets = self._graph.node_to_ref_offset[start_nodes]+start_offsets

            return FlatKmers(self._kmers.get_nparray(), self._nodes.get_nparray(), ref_offsets,
                             self._allele_frequencies.get_nparray())
        else:
            return FlatKmers2(self._kmers.get_nparray(), self._start_nodes.get_nparray(),
                          self._start_offsets.get_nparray(), self._nodes.get_nparray(),
                            self._allele_frequencies.get_nparray())

    def _add_kmer(self, kmer, start_node, start_offset):

        if self._whitelist is not None and kmer not in self._whitelist:
            self._n_skipped_whitelist += 1
            return

        nodes = np.unique(self._current_nodes[self._current_path_start_position:])
        #print("     Adding kmer %d/%s at node/offset %d/%d with nodes %s. Start node/offset: %d/%d" %
        #             (kmer, kmer_hash_to_sequence(kmer, self._k), start_node, start_offset, nodes, start_node, start_offset))

        n_variant_nodes = len([n for n in nodes if not self._graph.is_linear_ref_node_or_linear_ref_dummy_node(n)])
        if n_variant_nodes > self._max_variant_nodes:
            logging.warning("Passed too many variant nodes")


        kmer_allele_frequency = np.min(self._graph.get_node_allele_frequencies(nodes))

        if self._only_save_one_node_per_kmer:
            nodes = [nodes[0]]

        if len(self.kmers_found) < 500:
            nodes_added_set = set()
            
        # add one hit for each unique node
        for node in nodes:
            if self._only_store_nodes is not None and node not in self._only_store_nodes:
                continue

            #logging.info("    Adding start node %d for node %d" % (start_node, node))
            self._start_nodes.append(start_node)
            self._start_offsets.append(start_offset)
            self._nodes.append(node)
            self._kmers.append(kmer)
            self._allele_frequencies.append(kmer_allele_frequency)
            if len(self.kmers_found) < 500:
                nodes_added_set.add(node)

        if len(self.kmers_found) < 500:
            # Only add to this when there is little data, only used for testing and debugging
            # loggign.info("------Added kmer with nodes %s" % nodes)
            self.kmers_found.append((None, nodes_added_set, start_node, kmer))

    def find_only_kmers_starting_at_position(self, node, offset):
        # recuse all kmers from this position, always stop after a full kmer has been reached
        self._early_stop = True
        self._current_critical_node = node
        self._current_critical_offset = offset
        self._critical_graph_paths = CriticalGraphPaths.empty()
        #logging.info("Finding kmers starting at position %d/%d" % (node, offset))
        self.search_from(node, offset, 0)

    def find(self):
        if self._critical_graph_paths is None:
            logging.info("Making critical graph paths since it's not specified. "
                         "Will be faster if critical graph paths is premade")
            self._critical_graph_paths = CriticalGraphPaths.from_graph(self._graph, self._k)

        # iteratively calls search_from() on the various starting points (critical positions)

        logging.info("Stop at critical path number: %s" % self._stop_at_critical_path_number)
        logging.info("Start at critical path number: %s" % self._start_at_critical_path_number)

        self._starting_points = list(self._critical_graph_paths)[::-1]

        stop_at_node = None
        if self._stop_at_critical_path_number is not None and self._stop_at_critical_path_number < len(self._starting_points):
            stop_at_node = self._starting_points[-self._stop_at_critical_path_number-1][0]
            logging.info("Will stop at node %d" % stop_at_node)
        else:
            logging.info("Will stop at end of graph")


        self._starting_points_set = set()
        for starting_point in self._starting_points:
            self._starting_points_set.add(starting_point)

        if self._start_at_critical_path_number is not None and self._start_at_critical_path_number > 0:
            self._starting_points = self._starting_points[:-self._start_at_critical_path_number]  # remove the last

        # add beginning of graph as starting point if necessary
        if self._start_at_critical_path_number is None or self._start_at_critical_path_number == 0:
            if self._graph.get_node_size(self._graph.get_first_node()) <= self._k:  # means beginning is not a critical point, needs to add
                self._starting_points.append((self._graph.get_first_node(), 0))
                logging.info("Added first node of graph to critical nodes")


        #for critical_node, critical_offset in [(self._graph.get_first_node(), 0)] + list(self._critical_graph_paths):
        #logging.info("Starting points: %s" % self._starting_points)
        while len(self._starting_points) > 0:
            self._recursion_depth = 0
            critical_node, critical_offset = self._starting_points.pop()

            if stop_at_node is not None and stop_at_node == critical_node:
                logging.info("Stopping at critical path number %d" % self._stop_at_critical_path_number)
                break

            self._current_bases = NpList(dtype=np.int8)
            self._current_nodes = NpList()
            self._current_path_start_position = 0
            self._current_critical_node = critical_node
            self._current_critical_offset = critical_offset
            self._nonempty_bases_traversed = 0

            if critical_offset >= self._k-1:
                critical_offset -= (self._k-1)

            try:
                self.search_from(critical_node, critical_offset, 0)
            except RecursionError:
                logging.error("Failed searching from critical node %d and position %d" % (critical_node, critical_offset))
                logging.error("The graph might be too complex. Try setting max variants nodes lower?")
                logging.error("Recursion depth: %d" % self._recursion_depth)
                logging.error("Starting at critical path number: %d" % self._start_at_critical_path_number)
                raise

        logging.info("N nodes skipped because too many variant nodes: %d" % self._n_nodes_skipped_because_too_complex)
        logging.info("N skipped because whitelist: %d" % self._n_skipped_whitelist)


    def _is_critical_position(self, node, offset):
        # critical position means that there are only one kmer ending at this position (not multiple paths before)
        # and the node is a critical node
        if self._critical_graph_paths.is_critical(node, offset):
            return True
        return False

    def search_from(self, node, offset, current_hash):

        assert self._allele_frequencies._dtype == float, self._allele_frequencies._dtype

        self._recursion_depth += 1
        node_size = self._graph.get_node_size(node)

        if offset == 0 and self._graph.get_node_size(node) == 0:
            # this is a dummy node
            current_base = -1
            self._current_bases.append(current_base)
            self._current_nodes.append(node)


        while offset < node_size:

            # if we are in a middle of a big node, we can process this part more effectively since
            # we know there are no dummy nodes or edges or other stuff to think about
            if offset == self._k+2 and node_size > offset + self._k + 1 and not self._early_stop:
                current_hash, offset = self._process_whole_node(current_hash, node, node_size, offset)
            # change the current hash and current bases
            # if we don't have a long enough path, build hash incrementally by adding zero
            first_base = self._get_first_base_in_path()

            assert first_base != -1
            current_base = self._graph.get_numeric_base_sequence(node, offset)

            only_add = self._nonempty_bases_traversed  # len(self._current_bases)
            if current_base != -1:
                #if len(self._current_bases) >= self._k:
                if self._nonempty_bases_traversed >= self._k:
                    self._current_path_start_position += 1
                    only_add = False
                current_hash = update_hash(current_base, current_hash, first_base, self._k,
                                                                         only_add=only_add)
            if current_hash < 0:
                logging.error("Current node/offset: %d/%d" % (node, offset))
                logging.error("Current hash: %d" % current_hash)
                logging.error("First base: %d" % first_base)
                logging.error("Current base: %d" % current_base)
                raise Exception("Error in computing hash")

            self._current_bases.append(current_base)
            self._current_nodes.append(node)
            self._nonempty_bases_traversed += 1

            assert self._nonempty_bases_traversed <= len(self._current_bases)

            if False and ((node < 100 and offset < 15) or (offset > 0 and offset % 5000 == 0) or node % 5000 == 0):
                print("On node %d/%d, offset %d, %d kmers added. Skipped nodes: %d. "
                             "Path length: %d. Rec depth: %d. Nonempty traversed: %d. Nodes: %s"
                             % (node, len(self._graph.nodes), offset, len(self._kmers),
                                self._n_nodes_skipped_because_too_complex,
                                len(self._current_bases), self._recursion_depth, self._nonempty_bases_traversed,
                                ",".join((str(b) for b in (self._current_nodes[self._current_path_start_position:])))))
                #logging.info("Current search start node/offset: %d/%d" % (self._current_search_start_node, self._current_search_start_offset))

            current_path_desc = (node, offset, frozenset(self._current_nodes[self._current_path_start_position:]))
            if (node != self._current_critical_node or offset != self._current_critical_offset) and \
                    current_path_desc in self._positions_treated and len(self._current_nodes[self._current_path_start_position:]) >= self._k:
                self._recursion_depth -= 1
                #logging.info("Stopping recursion")
                #logging.info("current path desc: %s" % str((current_path_desc)))
                return False

            self._positions_treated.add(current_path_desc)

            # starts a depth first search from this position until meeting a long enough critical node in graph
            # do not add entries starting at empty nodes (we just want to include empty nodes in other entries)
            if self._nonempty_bases_traversed >= self._k and (current_base != -1 or self._early_stop):
                self._add_kmer(current_hash, node, offset)
                #logging.info("Added kmer")
                if self._early_stop:
                    #logging.info("Early stop")
                    # stop whenever a kmer is found
                    self._recursion_depth -= 1
                    return


            if (node != self._current_critical_node or offset + 1 != self._current_critical_offset) and \
                 self._is_critical_position(node, offset + 1):
                # stop recursion, next position is critical
                if (node, offset+1) not in self._starting_points_set:
                    self._starting_points.append((node, offset+1))
                    self._starting_points_set.add((node, offset+1))
                self._recursion_depth -= 1
                return False

            offset += 1

        assert offset == self._graph.get_node_size(node) or self._graph.get_node_size(node) == 0

        # at end of node, continue on edges
        self._search_next_nodes(current_hash, node)

    def _process_whole_node(self, current_hash, node, node_size, offset):
        sequence = self._graph.get_numeric_node_sequence(node)[offset - self._k:].astype(np.uint64)
        hashes = np.convolve(sequence, self._power_vector, mode='full')
        hashes = hashes[self._k:-self._k]  # get actual hashes after boundary effect and don't include last bp
        assert len(hashes) == node_size - offset - 1
        bases_to_extend = sequence[self._k:len(sequence) - 1]
        assert len(bases_to_extend) == len(hashes)
        self._current_bases.extend(bases_to_extend)
        self._current_nodes.extend(np.zeros(len(bases_to_extend)) + node)
        self._current_path_start_position += len(bases_to_extend)

        hashes_to_add = hashes
        offsets_to_add = np.arange(offset, node_size - 1)
        if self._whitelist is not None:
            hashes_to_add = np.array([h for h in hashes_to_add if h in self._whitelist])
            offsets_to_add = np.array([offset+i for i, h in enumerate(hashes) if h in self._whitelist])
            self._n_skipped_whitelist += (len(hashes)-len(hashes_to_add))

        n = len(hashes_to_add)


        self._kmers.extend(hashes_to_add)
        self._nodes.extend(np.zeros(n) + node)
        self._start_nodes.extend(np.zeros(n) + node)
        self._start_offsets.extend(offsets_to_add)
        self._allele_frequencies.extend(np.zeros(n, dtype=float) + self._graph.get_node_allele_frequency(node))

        # continue search from next offset and stop this search
        # NB: Converting to python int's to avoid problems when working with these hashes further
        current_hash = int(hashes[-1])
        # print("continuing with hash %d (type %s)" % (current_hash, type(current_hash)))
        offset = node_size - 1
        return current_hash, offset

    def _search_next_nodes(self, current_hash, node):
        next_nodes = self._graph.get_edges(node)
        force_follow = False
        if self._only_follow_nodes is not None and len(self._only_follow_nodes.intersection(next_nodes)) > 0:
            next_nodes = self._only_follow_nodes.intersection(next_nodes)
            force_follow = True

        if len(next_nodes) > 0:
            n_variant_nodes_passed = len(set([n for n in self._current_nodes[self._current_path_start_position:] if
                                              not self._graph.is_linear_ref_node_or_linear_ref_dummy_node(n)]))
            #logging.info("Searching next nodes from %d. Variant nodes until now: %d" % (node, n_variant_nodes_passed))
            if n_variant_nodes_passed > self._max_variant_nodes:
                logging.warning("Passed more variant nodes than planned. Could happen if forcing following certain nodes")

            if not force_follow and n_variant_nodes_passed >= self._max_variant_nodes:
                # only allow next nodes on linear ref
                self._n_nodes_skipped_because_too_complex += len(next_nodes)
                next_nodes = [node for node in next_nodes if
                              self._graph.is_linear_ref_node_or_linear_ref_dummy_node(node)]
                assert len(next_nodes) == 1, "Not 1 linear ref next nodes from node %d: %s" % (node, next_nodes)
                self._n_nodes_skipped_because_too_complex -= len(next_nodes)

            #logging.info("Next nodes: %s" % next_nodes)
            for i, next_node in enumerate(next_nodes):
                #logging.info("   Continuing on node %s. Nonempty bases: %d" % (next_node, self._nonempty_bases_traversed))
                path_start = self._current_path_start_position  # copy path start position before continuing recursion
                n_bases_in_path = len(self._current_bases)
                nonempty_bases_copy = self._nonempty_bases_traversed+0
                self.search_from(next_node, 0, current_hash)

                # after processing a child, reset current bases and nodes to where we are now before procesing next child
                self._current_bases.set_n_elements(n_bases_in_path)
                self._current_nodes.set_n_elements(n_bases_in_path)
                self._current_path_start_position = path_start
                self._nonempty_bases_traversed = nonempty_bases_copy

    def _get_first_base_in_path(self):
        #if len(self._current_bases) >= self._k:
        if self._nonempty_bases_traversed >= self._k:

            first_base = self._current_bases[self._current_path_start_position]

            # check if "dummy bases" are coming, then we want to skip them for the next iteration
            if len(self._current_bases) > self._current_path_start_position+1:
                next_first_base = self._current_bases[self._current_path_start_position + 1]
                while next_first_base == -1:
                    self._current_path_start_position += 1
                    next_first_base = self._current_bases[self._current_path_start_position + 1]

        else:
            first_base = 0
        return first_base



import logging
import time

from .snp_kmer_finder import SnpKmerFinder
from .flat_kmers import FlatKmers
from obgraph import VariantNotFoundException
from .kmer_finder import DenseKmerFinder


class UniqueVariantKmersFinder:
    def __init__(self, graph, variant_to_nodes, variants, k=31, max_variant_nodes=6,
                 kmer_index_with_frequencies=None, haplotype_matrix=None, node_to_variants=None,
                 do_not_choose_lowest_frequency_kmers=False, use_dense_kmer_finder=False, position_id_index=None,
                 use_simple=False):
        self.graph = graph
        self.variant_to_nodes = variant_to_nodes
        self.reference_kmer_index = None
        self.variants = variants
        self.k = k
        self.flat_kmers_found = []
        self.n_failed_variants = 0
        self._n_skipped_because_added_on_other_node = 0
        self._max_variant_nodes = max_variant_nodes
        self._kmer_index_with_frequencies = kmer_index_with_frequencies
        self.haplotype_matrix = haplotype_matrix
        self.node_to_variants = node_to_variants
        self._use_dense_kmer_finder = use_dense_kmer_finder
        self._position_id_index = position_id_index
        self._nodes_found = set()
        self._use_simple = use_simple

        if self._use_dense_kmer_finder:
            assert self._position_id_index is not None, "Position id index must be set when using dense kmer finder"

        self._choose_kmers_with_lowest_frequencies = True
        if do_not_choose_lowest_frequency_kmers:
            self._choose_kmers_with_lowest_frequencies = False

    def kmer_is_unique_on_reference_position(self, kmer, reference_position, ref_start, ref_end):
        # returns true if kmer is unique when ignoring kmers on same ref pos
        reference_kmers = self.reference_kmer_index.get_between(ref_start, ref_end)
        # Remove the one matching our position
        #logging.info("    Checking whether %d among reference kmers: %s" % (kmer, reference_kmers))
        for i, reference_kmer in enumerate(reference_kmers):
            pos = ref_start + i
            if pos != reference_position and reference_kmer == kmer:
                return False
        return True

    def _find_all_kmers_from_position_that_covers_node(self, chromosome, ref_position, cover_node):
        possible_ref_position_adjusted = self.graph.convert_chromosome_ref_offset_to_graph_ref_offset(
            ref_position, chromosome)
        finder = DenseKmerFinder(self.graph, self.k, None, position_id=self._position_id_index,
                                 max_variant_nodes=self._max_variant_nodes, only_store_nodes=set(cover_node),
                                 )
        position_node = self.graph.get_node_at_ref_offset(possible_ref_position_adjusted)
        position_offset = self.graph.get_node_offset_at_ref_offset(possible_ref_position_adjusted)
        finder.find_only_kmers_starting_at_position(node, offset)
        return finder.kmers_found()


    def _find_all_possible_kmer_sets_covering_node(self):
        pass


    def find_kmers_over_variant_node(self, variant, node):
        # tries to find the best possible kmers to represent node
        # if node is small, find only one kmer
        # if node is large, allow for multiple kmers
        # tries to choose kmers thare are rare

        start_pos = variant.position
        assert variant.type is not None, "Variant type must be set"
        if variant.type != "SNP":
            start_pos = start_pos + 1  # start pos for indels is 1 base before

        start_pos -= 1  # go from 1-based to 0-based coordinates

        node_sequence = self.graph.get_node_sequence(node)
        if node_sequence == "" or variant.type != "SNP":
            # empty node, we cannot start at node, start one node before
            # do this for both alleles if indel, so that we start from the same pos
            #logging.info("Finding start node by searching for ref offset %d at chromosomse %d" % (start_pos-1, variant.chromosome))
            start_node = self.graph.get_node_at_chromosome_and_chromosome_offset(variant.chromosome, start_pos-8)
            start_node_offset = self.graph.get_node_offset_at_chromosome_and_chromosome_offset(variant.chromosome, start_pos-8)
            #start_node_offset = self.graph.get_node_size(start_node)-
        else:
            start_node = node
            start_node_offset = 0

        finder = DenseKmerFinder(self.graph, self.k, None, position_id=self._position_id_index,
                                 max_variant_nodes=self._max_variant_nodes, only_store_nodes=set([node]),
                                 only_follow_nodes=set([node])
                                 )
        #logging.info("Searching from node %d offset %d for variant %s" % (start_node, start_node_offset, variant))
        finder.find_only_kmers_starting_at_position(start_node, start_node_offset)
        result = finder.get_flat_kmers(v="1")
        return result


    def find_kmers_over_structural_variant(self, variant, ref_node, variant_node):
        # find over ref node and variant node
        self.find_kmers_over_variant_node(variant, ref_node)
        self.find_kmers_over_variant_node(variant, variant_node)


    def find_kmers_over_variant(self, variant, ref_node, variant_node):
        # more simple version, does not try to find unique kmers
        ref_kmers = self.find_kmers_over_variant_node(variant, ref_node)
        variant_kmers = self.find_kmers_over_variant_node(variant, variant_node)
        return FlatKmers.from_multiple_flat_kmers([ref_kmers, variant_kmers])


    def find_unique_kmers_over_variant(self, variant, ref_node, variant_node):
        is_valid = False
        has_added = False
        # Start searching from before. The last position is probably the best for indels, since the sequence after is offset-ed.
        # We want to end on this, because this is what is chosen if all else fails
        possible_ref_positions = [variant.position - i for i in range(2, self.k - 2)][::4][::-1]
        valid_positions_found = []

        for possible_ref_position in possible_ref_positions:
            possible_ref_position_adjusted = self.graph.convert_chromosome_ref_offset_to_graph_ref_offset(possible_ref_position, variant.chromosome)
            is_valid = True
            only_store_nodes = set()
            for n in (ref_node, variant_node):
                if n not in self._nodes_found:
                    only_store_nodes.add(n)

            if not self._use_dense_kmer_finder:
                finder = SnpKmerFinder(self.graph, self.k, max_variant_nodes=self._max_variant_nodes, only_store_nodes=only_store_nodes,
                                       haplotype_matrix=self.haplotype_matrix, node_to_variants=self.node_to_variants, variant_to_nodes=self.variant_to_nodes)
                finder.find_kmers_from_linear_ref_position(possible_ref_position_adjusted)
            else:
                finder = DenseKmerFinder(self.graph, self.k, None, position_id=self._position_id_index,
                                         max_variant_nodes=self._max_variant_nodes, only_store_nodes=only_store_nodes,
                                         )
                node = self.graph.get_node_at_ref_offset(possible_ref_position_adjusted)
                offset = self.graph.get_node_offset_at_ref_offset(possible_ref_position_adjusted)
                finder.find_only_kmers_starting_at_position(node, offset)


            kmers_ref = set()
            kmers_variant = set()
            for kmer, nodes, ref_position, hash in finder.kmers_found:
                if ref_node in nodes:
                    kmers_ref.add(hash)
                if variant_node in nodes:
                    kmers_variant.add(hash)

                # not a requirement anymore
                #if not self.kmer_is_unique_on_reference_position(hash, ref_position, max(0, ref_position - 150), ref_position + 150 - self.k):
                #    is_valid = False
                #    #logging.info("  Found on linear ref elsewhere, %s" % kmer)
                #    break

                #if hash in self._hashes_added:
                #    is_valid = False
                #    self._n_skipped_because_added_on_other_node += 1

            if not is_valid:
                continue

            # Check if there are identical kmers on the two variant nodes
            if len(kmers_ref.intersection(kmers_variant)) > 0:
                #logging.info("    Felles kmer mellom ref og variant: %s / %s" % (kmers_ref, kmers_variant))
                is_valid = False

            # if we are at last position (then choose this anyway, better than nothing)
            if possible_ref_position == possible_ref_positions[-1]:
                is_valid = True
                #is_valid = False  # testing this

            if is_valid:
                # Kmers are valid, we don't need to search anymore for this variant
                flat = finder.get_flat_kmers(v="1")
                # hack for now: Set ref offset to be the ref offset we searched from
                # new densekmerfinder uses end as ref offset
                #flat._ref_offsets[:] = possible_ref_position_adjusted
                assert len(set(flat._nodes)) <= 2
                valid_positions_found.append(flat)

                if flat.maximum_kmer_frequency(self._kmer_index_with_frequencies) <= 1:
                    # no need to search for more positions, all kemrs are unique
                    break

                if len(flat._nodes) == 0:
                    logging.warning("Found 0 nodes for variant %s. Hashes: %s, ref positions: %s. Searched from ref position %d" % (variant, flat._hashes, flat._ref_offsets, possible_ref_position))
                    #raise Exception("No kmers found")

                if ref_node not in flat._nodes and ref_node in only_store_nodes:
                    logging.warning("Found kmers for variant %s with ref/variant nodes %d/%d but flat kmers does not contain ref node. Flat kmer nodes: %s. Searched from possition %d" % (variant, ref_node, variant_node, flat._nodes, possible_ref_position))
                    #raise Exception("No kmers found")

                if variant_node not in flat._nodes and variant_node in only_store_nodes:
                    logging.warning("No variant node kmers found for variant %s with variant node %d and ref node %d" %
                                    (variant, variant_node, ref_node))
                    logging.warning("Found no variant node kmers for variant %s. Hashes: %s, ref positions: %s. Searched from ref position %d" %
                                    (variant, flat._hashes, flat._ref_offsets, possible_ref_position))
                    logging.warning("Found nodes: %s" % flat._nodes)
                    logging.warning("Only store nodes: %s" % only_store_nodes)
                    logging.warning("Searched from node/offset: %d/%d" % (
                        self.graph.get_node_at_ref_offset(possible_ref_position),
                        self.graph.get_node_offset_at_ref_offset(possible_ref_position)
                    ))
                    #raise Exception("Errror")

        # Sort positions by max kmer frequency
        if len(valid_positions_found) == 0:
            logging.warning("Found no positions with valid kmers for variant %s" % variant)
            self.n_failed_variants += 1
            return

        if self._choose_kmers_with_lowest_frequencies:
            valid_positions_found = sorted(valid_positions_found, key=lambda p: p.maximum_kmer_frequency(self._kmer_index_with_frequencies))

        #valid_positions_found = sorted(valid_positions_found, key=lambda p: p.sum_of_kmer_frequencies(self._kmer_index_with_frequencies))
        best_position = valid_positions_found[0]
        nodes_found_here = list(best_position._nodes)
        if nodes_found_here.count(ref_node) > 1000 and False:
            logging.warning("Ref node %d found %d times" % (ref_node, nodes_found_here.count(ref_node)))
            logging.warning("Starting positions: %s" % possible_ref_positions)
            logging.info("Nodes: %s" % best_position._nodes)
            logging.info("Variant: %s" % variant)
            logging.info("Nodse from all positions:")
            for pos in valid_positions_found:
                logging.info(pos._nodes)

            assert False


        for node in set(best_position._nodes):
            assert node not in self._nodes_found, "Found node %d at variant %s (nodes %d/%d), but already found for previous variant" % \
                                                  (node, variant, ref_node, variant_node)
            self._nodes_found.add(node)

        return best_position
        #self.flat_kmers_found.append(best_position)


    def find_unique_kmers(self):
        prev_time = time.time()
        for i, variant in enumerate(self.variants):
            n_processed = len(self.flat_kmers_found)
            if i % 5000 == 0:
                logging.info("%d/%d variants processed (time spent on previous 5000 variants: %.3f s). "
                             "Now on chromosome/ref pos %s/%d" % (i, len(self.variants), time.time()-prev_time, str(variant.chromosome), variant.position))
                prev_time = time.time()

            #ref_node, variant_node = self.graph.get_variant_nodes(variant)
            assert variant.vcf_line_number is not None, "Variant line number must be specified"
            ref_node = self.variant_to_nodes.ref_nodes[variant.vcf_line_number]
            variant_node = self.variant_to_nodes.var_nodes[variant.vcf_line_number]

            if ref_node == 0 or variant_node == 0:
                continue

            if not self._use_simple:
                self.flat_kmers_found.append(self.find_unique_kmers_over_variant(variant, ref_node, variant_node))
            else:
                self.flat_kmers_found.append(self.find_kmers_over_variant(variant, ref_node, variant_node))

            if len(self.flat_kmers_found) != n_processed + 1:
                logging.warning("DID NOT FIND KMERS ON %s" % variant)

        logging.info("N variants with kmers found: %d" % len(self.flat_kmers_found))
        logging.info("Done with all variants. N that failed: %d" % self.n_failed_variants)

        return FlatKmers.from_multiple_flat_kmers(self.flat_kmers_found)




from npstructures.hashtable import HashTable
import numpy as np


class MultiValueHashTable:
    def __init__(self, hash_table, values):
        self._hash_table = hash_table
        self._values = values

    def get_unique_keys(self):
        return np.unique(self._hash_table._keys.ravel())

    def get_all_keys(self):
        return self._hash_table._keys.ravel()

    @classmethod
    def from_keys_and_values(cls, keys, values: dict, mod=None):
        hash_table = HashTable(keys, np.arange(len(keys), dtype=np.int64), mod=mod)
        return cls(hash_table, values)

    def __getitem__(self, keys):
        indexes = self._hash_table[keys]
        return {name: value[indexes] for name, value in self._values.items()}



import time
import numpy as np
import logging
import pickle
import gc

from graph_kmer_index import kmer_hash_to_sequence, sequence_to_kmer_hash
from npstructures import Counter, HashTable
from Bio.Seq import Seq
from .flat_kmers import FlatKmers

from .multi_value_hashtable import MultiValueHashTable

class CounterKmerIndex:
    def __init__(self, kmers, nodes, counter):
        self.kmers = kmers
        self.nodes = nodes
        self.counter = counter

    @classmethod
    def from_kmer_index(cls, kmer_index):
        kmers = kmer_index._kmers.astype(np.int64)
        nodes = kmer_index._nodes
        unique_kmers = np.unique(kmers)
        logging.info("N unique kmers: %d" % len(unique_kmers))
        #counter = Counter(unique_kmers, np.zeros_like(unique_kmers, dtype=np.uint16), mod=modulo, value_dtype=np.uint16)
        counter = Counter(unique_kmers, 0, mod=kmer_index._modulo, value_dtype=np.uint16)
        return cls(kmers, nodes, counter)

    def reset(self):
        self.counter = np.zeros_like(self.counter)

    def count_kmers(self, kmers, update_counter=True):
        if not update_counter:
            self.reset()

        self.counter.count(kmers.astype(np.int64))

    def get_node_counts(self, min_nodes=0):
        return np.bincount(self.nodes, self.counter[self.kmers], minlength=min_nodes)



class MinimalKmerIndex:
    def __init__(self, hashes_to_index, n_kmers, nodes, kmers, modulo):
        self._hashes_to_index = hashes_to_index.astype(np.int64)
        self._n_kmers = n_kmers.astype(np.uint32)
        self._nodes = nodes.astype(np.uint32)
        self._kmers = kmers
        self._modulo = np.int64(modulo)

    def max_node_id(self):
        return np.max(self._nodes)

    def to_file(self, file_name):
        logging.info("Writing kmer index to file: %s" % file_name)
        np.savez(file_name, hashes_to_index=self._hashes_to_index,
                 n_kmers=self._n_kmers,
                 nodes=self._nodes,
                 kmers=self._kmers,
                 modulo=self._modulo)

    @classmethod
    def from_file(cls, file_name):
        t = time.perf_counter()
        try:
            data = np.load(file_name + ".npz")
        except FileNotFoundError:
            data = np.load(file_name)

        logging.info("Took %.6f sec to read kmer index from file" % (time.perf_counter()-t))
        return cls(data["hashes_to_index"], data["n_kmers"], data["nodes"], data["kmers"], data["modulo"])


    @classmethod
    def from_flat_kmers(cls, flat_kmers, modulo=452930477):
        kmers = flat_kmers._hashes
        nodes = flat_kmers._nodes
        logging.info("Making hashes")
        hashes = kmers % modulo
        logging.info("Sorting")
        sorting = np.argsort(hashes)
        hashes = hashes[sorting]
        kmers = kmers[sorting]
        nodes = nodes[sorting]
        logging.info("Done sorting")

        # Find positions where hashes change (these are our index entries)
        diffs = np.ediff1d(hashes, to_begin=1)
        unique_entry_positions = np.nonzero(diffs)[0]
        try:
            unique_hashes = hashes[unique_entry_positions]
        except IndexError:
            logging.info("unique entry positions: %s" % unique_entry_positions)
            logging.info("Hashes: %s" % hashes)
            raise

        lookup = np.zeros(modulo, dtype=np.int)
        lookup[unique_hashes] = unique_entry_positions
        n_entries = np.ediff1d(unique_entry_positions, to_end=len(nodes) - unique_entry_positions[-1])
        n_kmers = np.zeros(modulo, dtype=np.uint32)
        n_kmers[unique_hashes] = n_entries

        # Find out how many entries there are for each unique hash
        object = cls(lookup, n_kmers, nodes, kmers, modulo)
        return object



class KmerIndex2:
    # same as CollisionFreeKmerIndex, but with startnode/offset in stead of ref_offset
    def __init__(self, data, frequencies=None):
        self._data = data
        self._frequencies = frequencies

    def get_start_nodes(self, kmer):
        return self._data[kmer]["start_nodes"]

    def get_start_offsets(self, kmer):
        return self._data[kmer]["start_offsets"]

    def get_nodes(self, kmer):
        return self._data[kmer]["nodes"]

    def get_all_kmers(self):
        return self._data.get_all_keys()


    def get_kmer_frequency(self, kmer):
        assert self._frequencies is not None, "Frequencies not set"
        return self._frequencies[kmer]

    @classmethod
    def from_flat_kmers(cls, flat_kmers, modulo=None, skip_frequencies=False):
        hash_table = MultiValueHashTable.from_keys_and_values(flat_kmers._hashes,
                                         {"nodes": flat_kmers._nodes,
                                          "start_nodes": flat_kmers._start_nodes,
                                          "start_offsets": flat_kmers._start_offsets,
                                          "allele_frequencies": flat_kmers._allele_frequencies},
                                         mod=modulo)

        index = cls(hash_table)
        if not skip_frequencies:
            index.count_unique_kmer_occurences()

        return index

    def count_unique_kmer_occurences(self):
        unique_kmers = self._data.get_unique_keys()
        counts = np.zeros_like(unique_kmers)

        # count unique entries (unique start node/offset) for each kmer
        for i, kmer in enumerate(unique_kmers):
            n = len(set((node, offset) for node, offset in
                        zip(self.get_start_nodes(kmer), self.get_start_offsets(kmer))))
            counts[i] = n

        self._frequencies = HashTable(unique_kmers, counts)




class CollisionFreeKmerIndex:
    properties = {
            "_hashes_to_index",
            "_n_kmers",
            "_nodes",
            "_ref_offsets",
            "_kmers",
            "_modulo",
            "_frequencies",
            "_allele_frequencies"
        }

    # ugly signature to support automatic reading from index bundle: TODO: Fix
    def __init__(self, _hashes_to_index=None, _n_kmers=None, _nodes=None, _ref_offsets=None, _kmers=None, _modulo=452930477, _frequencies=None, _allele_frequencies=None):
        self._hashes_to_index = _hashes_to_index
        self._n_kmers = _n_kmers
        self._nodes = _nodes
        self._ref_offsets = _ref_offsets
        self._kmers = _kmers  # Actual numeric kmers (not hashes of numeric kmers) at each position
                             # used to filter out collisions
        self._modulo = int(_modulo)
        if _frequencies is None:
            self._frequencies = 0
        else:
            self._frequencies = _frequencies

        self._allele_frequencies = _allele_frequencies

    def clear(self):
        self._hashes_to_index = None
        self._n_kmers = None
        self._nodes = None
        self._kmers = None
        self._modulo = None
        gc.collect()

    def copy(self):
        return CollisionFreeKmerIndex(self._hashes_to_index.copy(),
                                      self._n_kmers.copy(),
                                      self._nodes.copy(),
                                      self._ref_offsets.copy(),
                                      self._kmers.copy(),
                                      self._modulo.copy(),
                                      self._frequencies.copy(),
                                      self._allele_frequencies.copy()
                                      )

    def set_allele_frequencies(self, frequencies):
        pass

    def max_node_id(self):
        return np.max(self._nodes)

    def convert_to_int32(self):
        self._hashes_to_index = self._hashes_to_index.astype(np.int32)
        self._nodes = self._nodes.astype(np.int32)
        self._n_kmers = self._n_kmers.astype(np.int32)
        self._modulo = np.uint64(self._modulo)

    def remove_ref_offsets(self):
        self._ref_offsets = np.array([0])

    def set_frequencies_using_other_index(self, other, multiplier=1, min_frequency=1):
        unique = np.unique(self._kmers)
        for i, kmer in enumerate(unique):
            kmer = int(kmer)
            if i % 100000 == 0:
                logging.info("%d/%d unique kmers processed" % (i, len(unique)))
            frequency = other.get_frequency(kmer)
            hash = int(kmer) % self._modulo
            position = self._hashes_to_index[hash]
            n_hits = self._n_kmers[hash]
            start = position
            end = position + n_hits
            hit_positions = np.where(self._kmers[start:end] == kmer)[0]
            self._frequencies[hit_positions + start] = max(min_frequency, frequency * multiplier)

    def set_frequencies(self, skip=False):
        logging.info("Setting frequencies")
        # Count number of each kmer (not hashes, store these)
        self._frequencies = np.zeros(len(self._kmers), dtype=np.uint16)

        if skip:
            logging.info("Skipped setting frequencies. All frequencies are just 0 by default.")
            return

        unique = np.unique(self._kmers)
        for i, kmer in enumerate(unique):
            if i % 100000 == 0:
                logging.info("%d/%d unique kmers processed" % (i, len(unique)))

            hash = int(kmer) % self._modulo
            position = self._hashes_to_index[hash]
            n_hits = self._n_kmers[hash]
            start = position
            #assert start != 0 or hash == 0, "Kmer %d with hash %d, index position %d not found in index" % (kmer, hash, position)
            end = position + n_hits
            hit_positions = np.where(self._kmers[start:end] == kmer)[0]

            # The count is the number of unique ref positions here
            # (since same entry can have multiple nodes, but always same ref pos)
            count = len(set(self._ref_offsets[hit_positions + start]))
            assert count > 0, "Count is not > 0 for kmer %d, start, end: %d,%d. Ref offsets: %s" % (kmer, start, end, self._ref_offsets[hit_positions + start])
            self._frequencies[hit_positions + start] = count

    def __contains__(self, item):
        return self.get(int(item), 100000000000)[0] is not None


    def get_nodes(self, kmer, max_hits=10):
        nodes, offsets, frequencies, allele_frequencies = self.get(kmer, max_hits)
        return nodes

    def get(self, kmer, max_hits=10):
        hash = kmer % self._modulo
        position = self._hashes_to_index[int(hash)]
        n_hits = self._n_kmers[int(hash)]
        start = position
        end = position + n_hits
        hit_positions = np.where(self._kmers[start:end] == kmer)[0]
        frequencies = self._frequencies[hit_positions+start]
        allele_frequencies = self._allele_frequencies[hit_positions+start]
        if len(hit_positions) == 0 or frequencies[0] > max_hits:
            return None, None, None, None

        return self._nodes[hit_positions + start], self._ref_offsets[hit_positions + start], frequencies, allele_frequencies

    def get_grouped_nodes(self, kmer, max_hits=10):
        hits = self.get(kmer, max_hits)
        if hits[0] is None:
            return None

        ref_offsets = hits[1]
        nodes = hits[0]
        sorting = np.argsort(ref_offsets)
        ref_offsets = ref_offsets[sorting]
        nodes = nodes[sorting]


        _, hit_indexes = np.unique(ref_offsets, return_index=True)
        hit_indexes = list(hit_indexes)
        hit_indexes.append(len(ref_offsets))

        intervals = [(start, end) for start, end in zip(hit_indexes[0:-1], hit_indexes[1:])]
        return [nodes[start:end] for start, end in intervals]

    def get_frequency(self, kmer, include_reverse_complement=True, k=31):
        nodes, ref_offsets, frequencies, allele_frequencies = self.get(kmer, max_hits=1000000000000000)
        if nodes is None:
            f = 0
        else:
            f = int(frequencies[0])  # convert to avoid overflow error

        if include_reverse_complement:
            sequence = kmer_hash_to_sequence(kmer, k)
            rev_sequence = str(Seq(sequence).reverse_complement())
            rev_kmer = sequence_to_kmer_hash(rev_sequence)
            nodes, ref_offsets, frequencies, allele_frequencies = self.get(rev_kmer, max_hits=1000000000000000)

            if nodes is not None:
                f += int(frequencies[0])

        return f

    def get_nodes_and_ref_offsets_from_multiple_kmers(self, kmers, max_hits=10):
        all_nodes = []
        all_ref_offsets = []
        all_read_offsets = []
        all_frequencies = []
        for i, hash in enumerate(kmers):
            nodes, ref_offsets, frequencies, allele_frequencies = self.get(hash, max_hits=max_hits)
            if nodes is None:
                continue
            all_nodes.append(nodes)
            all_ref_offsets.append(ref_offsets)
            all_read_offsets.append(np.zeros(len(nodes)) + i)
            all_frequencies.append(frequencies)


        if len(all_nodes) == 0:
            return np.array([]), np.array([]), np.array([]), np.array([])

        all_nodes = np.concatenate(all_nodes)
        all_ref_offsets = np.concatenate(all_ref_offsets)
        all_read_offsets = np.concatenate(all_read_offsets)
        all_frequencies = np.concatenate(all_frequencies)
        return all_nodes, all_ref_offsets, all_read_offsets, all_frequencies

    def get_nodes_from_multiple_kmers(self, kmers, max_hits=10):
        all_nodes = []
        for i, hash in enumerate(kmers):
            nodes, ref_offsets, frequencies, allele_frequencies = self.get(hash, max_hits=max_hits)
            if nodes is None:
                continue
            all_nodes.append(nodes)


        if len(all_nodes) == 0:
            return np.array([])

        all_nodes = np.concatenate(all_nodes)
        return all_nodes

    def to_file(self, file_name):
        logging.info("Writing kmer index to file: %s" % file_name)
        np.savez(file_name, hashes_to_index=self._hashes_to_index,
                 n_kmers=self._n_kmers,
                 nodes=self._nodes,
                 ref_offsets=self._ref_offsets,
                 kmers=self._kmers,
                 modulo=self._modulo,
                 frequencies=self._frequencies,
                 allele_frequencies=self._allele_frequencies)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name + ".npz")
        except FileNotFoundError:
            data = np.load(file_name)

        try:
            if "allele_frequencies" in data:
                allele_frequencies = data["allele_frequencies"]
            else:
                allele_frequencies = np.zeros(len(data["ref_offsets"]))

            return cls(data["hashes_to_index"], data["n_kmers"], data["nodes"], data["ref_offsets"], data["kmers"], data["modulo"], data["frequencies"], allele_frequencies)
        except AttributeError:
            logging.info("Keys are %s" % (data.keys()))
            raise

    @classmethod
    def from_flat_kmers(cls, flat_kmers, modulo=452930477, skip_frequencies=False, skip_singletons=False):
        if skip_singletons:
            flat_kmers = flat_kmers.get_new_without_singletons()

        kmers = flat_kmers._hashes
        nodes = flat_kmers._nodes
        ref_offsets = flat_kmers._ref_offsets

        logging.info("Making hashes")
        hashes = kmers % modulo
        logging.info("Sorting")
        sorting = np.argsort(hashes)
        hashes = hashes[sorting]
        kmers = kmers[sorting]
        nodes = nodes[sorting]
        ref_offsets = ref_offsets[sorting]
        allele_frequencies = flat_kmers._allele_frequencies[sorting]
        logging.info("Done sorting")

        # Find positions where hashes change (these are our index entries)
        diffs = np.ediff1d(hashes, to_begin=1)
        unique_entry_positions = np.nonzero(diffs)[0]
        try:
            unique_hashes = hashes[unique_entry_positions]
        except IndexError:
            logging.info("unique entry positions: %s" % unique_entry_positions)
            logging.info("Hashes: %s" % hashes)
            raise

        lookup = np.zeros(modulo, dtype=np.int32)
        lookup[unique_hashes] = unique_entry_positions
        n_entries = np.ediff1d(unique_entry_positions, to_end=len(nodes)-unique_entry_positions[-1])
        n_kmers = np.zeros(modulo, dtype=np.uint32)
        n_kmers[unique_hashes] = n_entries

        # Find out how many entries there are for each unique hash
        object = cls(lookup, n_kmers, nodes, ref_offsets, kmers, modulo, _allele_frequencies=allele_frequencies)
        object.set_frequencies(skip_frequencies)

        if skip_singletons:
            logging.info("Adding 1 to all frequencies since singletons are skipped")
            object._frequencies += 1

        return object


    def convert_kmers_to_complement(self, k=31, skip_frequencies=True):
        from .kmer_hashing import kmer_hashes_to_complement_hashes
        new_kmers = []
        for i, kmer_chunk in enumerate(np.array_split(self._kmers, len(self._kmers)//10000000)):
            logging.info("Doing chunk %d" % i)
            new_kmers.append(kmer_hashes_to_complement_hashes(kmer_chunk, k))

        logging.info("Concatenating")
        new_kmers = np.concatenate(new_kmers)

        logging.info("Building")
        return CollisionFreeKmerIndex.from_flat_kmers(
            FlatKmers(
                new_kmers,
                self._nodes,
                self._ref_offsets,
                self._allele_frequencies,
            ),
            modulo=self._modulo,
            skip_frequencies=skip_frequencies
        )





import logging
import numpy as np
from obgraph.variant_to_nodes import VariantToNodes
from obgraph.numpy_variants import NumpyVariants
from graph_kmer_index import KmerIndex
from shared_memory_wrapper import from_file, to_file

# Wrapper for all indexes required by kage genotyping
class IndexBundle:
    index_names = ["VariantToNodes", "NumpyVariants", "NodeCountModelAdvanced", "HelperVariants", "CombinationMatrix", "TrickyVariants", "KmerIndex"]
    def __init__(self, indexes):
        self.indexes = indexes

    @classmethod
    def from_file(cls, file_name, skip=None):
        return cls(from_file(file_name))

    def to_file(self, file_name, compress=True):
        return to_file(self.indexes, file_name, compress=compress)



from npstructures.hashtable import HashTable
import numpy as np


class NpSet:
    def __init__(self, values):
        if isinstance(values, HashTable):
            self._values = values
        else:
            unique_values = np.unique(values)
            self._values = HashTable(unique_values, np.ones(len(unique_values)))

    def __len__(self):
        return self

    def __contains__(self, item):
        return len(self._values[self._values.dtype(item)]) > 0




import itertools
import logging
import numpy as np
from graph_kmer_index import kmer_to_hash_fast, letter_sequence_to_numeric
from Bio.Seq import Seq
from .kmer_hashing import power_array


class ReadKmers:
    def __init__(self, kmers):
        self.kmers = kmers
        self._power_vector = None

    @classmethod
    def from_fasta_file(cls, fasta_file_name, k, small_k=None, smallest_k=8):
        power_vector = power_array(k)
        f = open(fasta_file_name)
        f = [l for l in f.readlines() if not l.startswith(">")]
        logging.info("Number of lines: %d" % len(f))
        if small_k is None:
            kmers = itertools.chain(
                (ReadKmers.get_kmers_from_read_dynamic(line.strip(), power_vector)
                        for line in f if not line.startswith(">")),
                (ReadKmers.get_kmers_from_read_dynamic(str(Seq(line.strip()).reverse_complement()), power_vector)
                        for line in f if not line.startswith(">"))
            )
        else:
            power_vector_small = power_array(small_k)
            power_vector_smallest = power_array(smallest_k)
            kmers = zip(
                    (itertools.chain(

                        ReadKmers.get_kmers_from_read_dynamic(line.strip(), power_vector),
                        ReadKmers.get_kmers_from_read_dynamic(str(Seq(line.strip()).reverse_complement()), power_vector)

                    ) for line in f),
                    (itertools.chain(
                            ReadKmers.get_kmers_from_read_dynamic(line.strip(), power_vector_small),
                            ReadKmers.get_kmers_from_read_dynamic(str(Seq(line.strip()).reverse_complement()), power_vector_small)
                    )
                    for line in f),
                    (itertools.chain(
                        ReadKmers.get_kmers_from_read_dynamic(line.strip(), power_vector_smallest),
                        ReadKmers.get_kmers_from_read_dynamic(str(Seq(line.strip()).reverse_complement()), power_vector_smallest)
                    )
                    for line in f)
            )

        return cls(kmers)

    @classmethod
    def from_list_of_string_kmers(cls, string_kmers):
        kmers = [
            [kmer_to_hash_fast(letter_sequence_to_numeric(k), len(k)) for k in read_kmers]
            for read_kmers in string_kmers
        ]
        return cls(kmers)

    @staticmethod
    def get_kmers_from_read(read, k):
        kmers = []
        for i in range(len(read) - k):
            letter_sequence = letter_sequence_to_numeric(read[i:i+k])
            kmers.append(kmer_to_hash_fast(letter_sequence, k))
        return kmers

    @staticmethod
    def get_kmers_from_read_dynamic(read, power_vector):
        read = letter_sequence_to_numeric(read)
        return np.convolve(read, power_vector, mode='valid')

    @staticmethod
    def get_kmers_from_read_dynamic_slow(read, k):
        raise NotImplementedError()
        read = letter_sequence_to_numeric(read)
        kmers = np.zeros(len(read)-k+1, dtype=np.int64)
        current_hash = kmer_to_hash_fast(read[0:k], k)
        kmers[0] = current_hash
        for i in range(1, len(read)-k+1):
            kmers[i] = (kmers[i-1] - np.power(4, k-1) * read[i-1]) * 4 + read[i+k-1]
            #assert kmers[i] == kmer_to_hash_fast(read[i:i+k], k), "New hash %d != correct %d" % (kmers[i], kmer_to_hash_fast(read[i:i+k], k))

        return kmers

    def __iter__(self):
        return self.kmers.__iter__()

    def __next__(self):
        return self.kmers.__next__()


import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')
import sys
import argparse
import itertools
#from multiprocessing import shared_memory
import time
from multiprocessing import Pool, Process
import numpy as np
from itertools import repeat
from pyfaidx import Fasta

from .collision_free_kmer_index import CollisionFreeKmerIndex

from obgraph import Graph
from .snp_kmer_finder import SnpKmerFinder
import pickle
from .flat_kmers import FlatKmers
from .reverse_kmer_index import ReverseKmerIndex
from .unique_kmer_index import UniqueKmerIndex
from .reference_kmer_index import ReferenceKmerIndex
from pathos.multiprocessing import Pool
from obgraph.variants import VcfVariants
from .unique_variant_kmers import UniqueVariantKmersFinder
from shared_memory_wrapper.shared_memory import to_shared_memory, from_shared_memory, remove_shared_memory_in_session, to_file, from_file, object_to_shared_memory, object_from_shared_memory
from shared_memory_wrapper import get_shared_pool, close_shared_pool
from obgraph.variant_to_nodes import VariantToNodes, NodeToVariants
from obgraph.haplotype_matrix import HaplotypeMatrix
from npstructures import HashTable, Counter
from .npset import NpSet
from .collision_free_kmer_index import CounterKmerIndex
from .kmer_finder import DenseKmerFinder
from .collision_free_kmer_index import KmerIndex2
from .critical_graph_paths import CriticalGraphPaths
from shared_memory_wrapper import to_file, from_file
from obgraph.position_id import PositionId



def main():
    run_argument_parser(sys.argv[1:])


def create_index_single_thread(args, interval=None):
    start_position = None
    end_position = None
    if interval is not None:
        start_position = interval[0]
        end_position = interval[1]

    logging.info("Loading data")
    #graph = Graph.from_file(args.graph_file_name)
    if args.graph_file_name is not None:
        graph = from_shared_memory(Graph, "graph_shared"+args.shared_memory_unique_id)
        reference = None
    else:
        graph = None
        assert args.reference_fasta is not None
        assert args.reference_name is not None, "Reference name must be specified"
        logging.info("Reference name is *%s*" % args.reference_name)
        try:
            fasta = Fasta(args.reference_fasta)
            logging.info("Names in fasta: %s" % str(fasta.keys()))
            reference = fasta[args.reference_name]
            assert len(reference) > 0, "Length of ref sequennce is 0. Seomthing is wrong"
        except KeyError:
            logging.error("Did not find reference name %s in %s" % (args.reference_name, args.reference_fasta))
            raise

    logging.info("Running kmerfinder")
    whitelist = None
    if args.whitelist is not None:
        w = FlatKmers.from_file(args.whitelist)
        whitelist = set(w._hashes)

    skip_kmers_with_nodes = None
    if args.skip_kmers_with_nodes is not None:
        f = FlatKmers.from_file(args.skip_kmers_with_nodes)
        skip_kmers_with_nodes = set(f._nodes)

    finder = SnpKmerFinder(graph, k=args.kmer_size, spacing=args.spacing,
                           include_reverse_complements=False,  # always false, adds revcomp in the end if chosen
                           pruning=args.pruning,
                           max_kmers_same_position=args.max_kmers_same_position,
                           max_frequency=args.max_frequency,
                           max_variant_nodes=args.max_variant_nodes,
                           only_add_variant_kmers=args.only_add_variant_kmers,
                           whitelist=whitelist,
                           only_save_variant_nodes=args.only_save_variant_nodes,
                           start_position=start_position,
                           end_position=end_position,
                           skip_kmers_with_nodes=skip_kmers_with_nodes,
                           only_save_one_node_per_kmer=args.only_save_one_node_per_kmer,
                           reference=reference)

    kmers = finder.find_kmers()

    if args.include_reverse_complement:
        logging.info("Adding reverse complements")
        kmers_revcomp = kmers.get_reverse_complement_flat_kmers(args.kmer_size)
        kmers = FlatKmers.from_multiple_flat_kmers([kmers, kmers_revcomp])

    return kmers

def create_index(args):
    args.shared_memory_unique_id = str(np.random.randint(0, 10e15))
    r = args.shared_memory_unique_id

    if args.graph_file_name is not None:
        graph = Graph.from_file(args.graph_file_name)
        to_shared_memory(graph, "graph_shared"+r)

    if args.threads == 1:
        kmers = create_index_single_thread(args)
        kmers.to_file(args.out_file_name)
    else:
        n_jobs = args.threads * 10
        logging.info("Making pool with %d workers" % n_jobs)
        pool = Pool(args.threads)
        genome_size = args.genome_size
        n_total_start_positions = genome_size // args.spacing
        n_positions_each_process = n_total_start_positions // n_jobs
        logging.info("Using genome size %d. Will process %d genome positions in each process." % (genome_size, n_positions_each_process))
        intervals = []
        for i in range(n_jobs):
            start_position = n_positions_each_process * i * args.spacing
            end_position = n_positions_each_process * (i+1) * args.spacing
            intervals.append((start_position, end_position))
            logging.info("Creating interval for genome segment %d-%d" % (start_position, end_position))

        all_hashes = []
        all_nodes = []
        all_ref_offsets = []
        all_allele_frequencies = []
        for flat_kmers in pool.starmap(create_index_single_thread, zip(repeat(args), intervals)):
            all_hashes.append(flat_kmers._hashes)
            all_nodes.append(flat_kmers._nodes)
            all_ref_offsets.append(flat_kmers._ref_offsets)
            all_allele_frequencies.append(flat_kmers._allele_frequencies)

        logging.info("Making full index from all indexes")
        full_index = FlatKmers(
            np.concatenate(all_hashes),
            np.concatenate(all_nodes),
            np.concatenate(all_ref_offsets),
            np.concatenate(all_allele_frequencies)
        )

        logging.info("Saving full index")
        full_index.to_file(args.out_file_name)


def make_from_flat(args):
    flat = FlatKmers.from_file(args.flat_index)

    if args.add_reverse_complements:
        logging.info("Will add reverse complements of every hash. Getting reverse complements.")
        logging.info("NB: Using k=%d when computing reverse complements" % args.kmer_size)
        rev_comp_flat = flat.get_reverse_complement_flat_kmers(k=args.kmer_size)
        flat = FlatKmers.from_multiple_flat_kmers([flat, rev_comp_flat])
        logging.info("Made new flat kmers with rev comp, now making index")

    if args.make_minimal:
        logging.info("Making minimal kmer index")
        from .collision_free_kmer_index import MinimalKmerIndex
        index = MinimalKmerIndex.from_flat_kmers(flat, modulo=args.hash_modulo)
    else:
        index = CollisionFreeKmerIndex.from_flat_kmers(flat, modulo=args.hash_modulo, skip_frequencies=args.skip_frequencies,
                                                   skip_singletons=args.skip_singletons)
    index.to_file(args.out_file_name)
    logging.info("Done making kmer index")


def make_reverse(args):
    flat = FlatKmers.from_file(args.flat_index)
    reverse = ReverseKmerIndex.from_flat_kmers(flat)
    reverse.to_file(args.out_file_name)
    logging.info("Done. Wrote reverse index to file: %s" % args.out_file_name)


def make_reference_kmer_index(args):
    if args.reference_fasta is not None:
        logging.info("Making from a linear reference")
        index = ReferenceKmerIndex.from_linear_reference(args.reference_fasta, args.reference_name, args.kmer_size, args.only_store_kmers)
    else:
        flat = FlatKmers.from_file(args.flat_index)
        index = ReferenceKmerIndex.from_flat_kmers(flat)

    index.to_file(args.out_file_name)
    logging.info("Saved reference kmer index to file %s" % args.out_file_name)


def make_unique_index(args):
    graph = Graph.from_file(args.graph)
    reverse = ReverseKmerIndex.from_file(args.reverse)
    flat = FlatKmers.from_file(args.flat_index)
    unique = UniqueKmerIndex.from_flat_kmers_and_snps_graph(flat, graph, reverse)
    unique.to_file(args.out_file_name)


def prune_flat_kmers(args):
    index = FlatKmers.from_file(args.flat_index)
    new_hashes = []
    new_nodes = []
    new_ref_offsets = []

    prev_hash = -1
    prev_ref_offset = -1
    n_skipped = 0
    for i in range(len(index._hashes)):
        if i % 1000000 == 0:
            logging.info("%d processed, %d skipped" % (i, n_skipped))

        if index._hashes[i] == prev_hash and index._ref_offsets[i] == prev_ref_offset:
            n_skipped += 1
            continue


        new_hashes.append(index._hashes[i])
        new_nodes.append(index._nodes[i])
        new_ref_offsets.append(index._ref_offsets[i])

        prev_hash = index._hashes[i]
        prev_ref_offset = index._ref_offsets[i]

    new = FlatKmers(
        np.array(new_hashes, dtype=index._hashes.dtype),
        np.array(new_nodes, dtype=index._nodes.dtype),
        np.array(new_ref_offsets, dtype=index._ref_offsets.dtype),
    )




def run_argument_parser(args):
    parser = argparse.ArgumentParser(
        description='Graph Kmer Index.',
        prog='graph_kmer_index',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser("make")
    subparser.add_argument("-g", "--graph_file_name", required=False)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-k", "--kmer_size", required=False, type=int, default=31)
    subparser.add_argument("-r", "--include-reverse-complement", required=False, type=bool, default=False)
    subparser.add_argument("-s", "--spacing", required=False, type=int, default=31)
    subparser.add_argument("-p", "--pruning", required=False, type=bool, default=False, help="Set to True to prune unecessary kmers")
    subparser.add_argument("-m", "--max-kmers-same-position", required=False, type=int, default=100000, help="Maximum number of kmers allowd to be added from the same ref position")
    subparser.add_argument("-M", "--max-frequency", required=False, type=int, default=10000000, help="Skip kmers with frequency higher than this. Will never skip kmers crossing variants.")
    subparser.add_argument("-v", "--max-variant-nodes", required=False, type=int, default=100000, help="Max variant nodes allowed in kmer.")
    subparser.add_argument("-V", "--only-add-variant-kmers", required=False, type=bool, default=False)
    subparser.add_argument("-N", "--only-save-variant-nodes", required=False, type=bool, default=False)
    subparser.add_argument("-O", "--only-save-one-node-per-kmer", required=False, type=bool, default=False)
    subparser.add_argument("-S", "--skip-kmers-with-nodes", required=False, help="Skip kmers with nodes that exist in this flat kmers file")
    subparser.add_argument("-w", "--whitelist", required=False, help="Only add kmers in this whitelist (should be a flat kmers file)")
    subparser.add_argument("-t", "--threads", required=False, default=1, type=int, help="How many threads to use. Some parameters will have local effect if t > 1 (-M)")
    subparser.add_argument("-G", "--genome-size", required=False, default=3000000000, type=int, help="Must be set if --threads > 1 (used to make chunks to run in parallel)")
    subparser.add_argument("-R", "--reference-fasta", required=False, help="Make from this reference fasta instead of graph")
    subparser.add_argument("-n", "--reference-name", required=False, help="Name of reference in fasta file. Needed when reference fasta is used.")

    subparser.set_defaults(func=create_index)

    subparser = subparsers.add_parser("make_from_flat")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-f", "--flat-index", required=True)
    subparser.add_argument("-m", "--hash_modulo", required=False, type=int, default=452930477)
    subparser.add_argument("-S", "--skip-frequencies", type=bool, default=False, required=False)
    subparser.add_argument("-s", "--skip-singletons", type=bool, default=False, required=False)
    subparser.add_argument("-M", "--make-minimal", type=bool, default=False, required=False)
    subparser.add_argument("-r", "--add-reverse-complements", type=bool, default=False, required=False, help="If True, the rev. comp. of every kmer hash will also be added")
    subparser.add_argument("-k", "--kmer-size", type=int, default=31, required=False, help="Only required when --add-reverse-complements=True. Used to compute rev comp hashes.")
    subparser.set_defaults(func=make_from_flat)

    subparser = subparsers.add_parser("make_reverse")
    subparser.add_argument("-f", "--flat-index", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=make_reverse)

    subparser = subparsers.add_parser("make_unique_index", help="Make index from unique kmers in the graph to nodes that are covered by those kmers")
    subparser.add_argument("-f", "--flat-index", required=True)
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-r", "--reverse", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=make_unique_index)

    subparser = subparsers.add_parser("make_reference_kmer_index", help="Make index that allows lookup between two ref positions")
    subparser.add_argument("-f", "--flat-index", required=False, help="If set, will create from flat kmers")
    subparser.add_argument("-r", "--reference-fasta", required=False, help="If set, will create from a linear reference")
    subparser.add_argument("-n", "--reference-name", required=False, help="Only needed if creating from linear reference")
    subparser.add_argument("-k", "--kmer-size", required=False, type=int, default=16, help="Only needed if making from linear fasta")
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-O", "--only-store-kmers", required=False, default=False, type=bool, help="Can be used when making from fasta file, will then not store the index since an index is not needed")
    subparser.set_defaults(func=make_reference_kmer_index)

    def make_unique_variant_kmers_single_thread(variants, args):
        args = object_from_shared_memory(args)
        #variant_to_nodes = from_shared_memory(VariantToNodes, "variant_to_nodes_shared"+r)
        variant_to_nodes = args["variant_to_nodes"]

        #kmer_index = from_shared_memory(CollisionFreeKmerIndex, "kmer_index_shared"+r)
        kmer_index = args["kmer_index"]
        if kmer_index is None:
            kmer_index = args["kmer_counter"]

        #graph = from_shared_memory(Graph, "graph_shared"+r)
        graph = args["graph"]

        #graph = Graph.from_file(args.graph)
        logging.info("Reading all variants")

        node_to_variants = None
        haplotype_matrix = None
        """
        if args.haplotype_matrix is not None:
            haplotype_matrix = HaplotypeMatrix.from_file(args.haplotype_matrix)
            node_to_variants = NodeToVariants.from_file(args.node_to_variants)
        """

        use_dense_kmer_finder = False
        if args["use_dense_kmer_finder"]:
            use_dense_kmer_finder = True
            logging.info("Will use dense kmer finder")

        finder = UniqueVariantKmersFinder(graph, variant_to_nodes, variants, args["kmer_size"], args["max_variant_nodes"],
                                          kmer_index_with_frequencies=kmer_index, haplotype_matrix=haplotype_matrix,
                                          node_to_variants=node_to_variants,
                                          do_not_choose_lowest_frequency_kmers=args["do_not_choose_lowest_frequency_kmers"],
                                          use_dense_kmer_finder=use_dense_kmer_finder,
                                          position_id_index=args["position_id_index"],
                                          use_simple=args["simple"]
                                          )
        flat_kmers = finder.find_unique_kmers()
        return flat_kmers

    def make_unique_variant_kmers(args):
        args = vars(args)
        args_orig = args
        args.pop("func")  # necessary for putting args in shared memory

        logging.info("Reading all variants")
        variants = VcfVariants.from_vcf(args["vcf"], skip_index=True, make_generator=True)
        variants = variants.get_chunks(chunk_size=args["chunk_size"])
        pool = Pool(args["n_threads"])

        all_flat_kmers = []
        args = object_to_shared_memory(args)
        for flat_kmers in pool.starmap(make_unique_variant_kmers_single_thread, zip(variants, repeat(args))):
            all_flat_kmers.append(flat_kmers)

        logging.info("Merge all flat kmers")
        merged_flat = FlatKmers.from_multiple_flat_kmers(all_flat_kmers)
        merged_flat.to_file(args_orig["out_file_name"])
        logging.info("Wrote to file %s" % args_orig["out_file_name"])

    subparser = subparsers.add_parser("make_unique_variant_kmers", help="Make a reverse variant index lookup to unique kmers on that variant")
    subparser.add_argument("-g", "--graph", required=True, type=Graph.from_file)
    subparser.add_argument("-V", "--variant_to_nodes", required=True, type=VariantToNodes.from_file)
    subparser.add_argument("-N", "--node-to-variants", required=False)
    subparser.add_argument("-H", "--haplotype-matrix", required=False)
    subparser.add_argument("-k", "--kmer-size", required=True, type=int)
    subparser.add_argument("-i", "--kmer-index", required=False, help="Kmer index used to check frequency of kmers in genome", type=CollisionFreeKmerIndex.from_file)
    subparser.add_argument("-I", "--kmer-counter", required=False, help="Kmer index used to check frequency of kmers in genome", type=from_file)
    subparser.add_argument("-p", "--position-id-index", required=False, type=from_file)
    subparser.add_argument("-D", "--use-dense-kmer-finder", required=False, type=bool, default=False)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-v", "--vcf", required=True)
    subparser.add_argument("-t", "--n-threads", required=False, default=1, type=int)
    subparser.add_argument("-c", "--chunk-size", required=False, default=10000, type=int, help="Number of variants given to each thread")
    subparser.add_argument("-m", "--max-variant-nodes", required=False, default=6, type=int, help="Maximum number of variant nodes allowed in kmer")
    subparser.add_argument("-d", "--do-not-choose-lowest-frequency-kmers", required=False, type=bool, help="For testing only. Will not choose the best kmers.")
    subparser.add_argument("-S", "--simple", type=bool, default=False, help="Set to True to use simple kmer selection")
    subparser.set_defaults(func=make_unique_variant_kmers)


    def sample_kmers_from_structural_variants_command(args):
        from .structural_variants import sample_kmers_from_structural_variants
        kmers = sample_kmers_from_structural_variants(args.graph,
                                                      args.variant_to_nodes,
                                                      args.kmer_index,
                                                      args.kmer_size)
        kmers.to_file(args.out_file_name)


    subparser = subparsers.add_parser("sample_kmers_from_structural_variants", help="Samples extra kmers from large structural variant nodes. Meant ot be combined with other variant kmers")
    subparser.add_argument("-g", "--graph", required=True, type=Graph.from_file)
    subparser.add_argument("-V", "--variant_to_nodes", required=True, type=VariantToNodes.from_file)
    subparser.add_argument("-k", "--kmer-size", required=True, type=int)
    subparser.add_argument("-i", "--kmer-index", required=False,
                           help="Kmer index used to check frequency of kmers in genome",
                           type=from_file)
    subparser.add_argument("-I", "--kmer-counter", required=False,
                           help="Kmer index used to check frequency of kmers in genome", type=from_file)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-t", "--n-threads", required=False, default=1, type=int)
    subparser.set_defaults(func=sample_kmers_from_structural_variants_command)


    subparser = subparsers.add_parser("prune_flat_kmers")
    subparser.add_argument("-f", "--flat-index", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=prune_flat_kmers)

    def merge_flat_kmers(args):
        new = FlatKmers.from_multiple_flat_kmers([FlatKmers.from_file(f) for f in args.flat_kmers.split(",")])
        new.to_file(args.out_file_name)
        logging.info("Wrote merged index to %s" % new)

    subparser = subparsers.add_parser("merge_flat_kmers")
    subparser.add_argument("-f", "--flat-kmers", required="true", help="Comma-separeted list of file names to be merged")
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=merge_flat_kmers)

    def make_kmer_frequencies(args):
        from .kmer_frequency_index import KmerFrequencyIndex
        ref_kmers = ReferenceKmerIndex.from_file(args.reference_kmers)
        index = KmerFrequencyIndex.from_kmers(ref_kmers.kmers)
        index.to_file(args.out_file_name)
        logging.info("Wrote to file %s" % args.out_file_name)

    subparser = subparsers.add_parser("make_kmer_frequency_index")
    subparser.add_argument("-r", "--reference-kmers", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=make_kmer_frequencies)

    def set_frequencies_using_other_index(args):

        logging.info("Reading index")
        index = CollisionFreeKmerIndex.from_file(args.kmer_index)
        logging.info("Reading other index")
        other = CollisionFreeKmerIndex.from_file(args.kmer_index_with_frequencies)
        index.set_frequencies_using_other_index(other, args.multiplier)
        index.to_file(args.kmer_index)
        logging.info("Wrote index to file %s" % args.kmer_index)

    subparser = subparsers.add_parser("set_frequencies_using_other_index")
    subparser.add_argument("-i", "--kmer-index", required=True)
    subparser.add_argument("-f", "--kmer-index-with-frequencies", required=True)
    subparser.add_argument("-m", "--multiplier", required=False, type=int, default=1)
    subparser.set_defaults(func=set_frequencies_using_other_index)

    def set_allele_frequencies(args):
        index = CollisionFreeKmerIndex.from_file(args.kmer_index)
        frequencies = np.load(args.frequencies)
        index.set_allele_frequencies(frequencies)
        index.to_file(args.kmer_index)
        logging.info("Wrote index to file %s" % args.kmer_index)

    subparser = subparsers.add_parser("set_allele_frequencies")
    subparser.add_argument("-i", "--kmer-index", required=True)
    subparser.add_argument("-f", "--frequencies", required=True)
    subparser.set_defaults(func=set_allele_frequencies)


    def create_counter_index(args):
        kmer_index = CollisionFreeKmerIndex.from_file(args.kmer_index)
        counter_index = CounterKmerIndex.from_kmer_index(kmer_index, args.modulo)
        to_file(counter_index, args.out_file_name)


    subparser = subparsers.add_parser("create_counter_index")
    subparser.add_argument("-i", "--kmer-index", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-m", "--modulo", required=False, default=200000033, type=int)
    subparser.set_defaults(func=create_counter_index)


    def index_single_thread(data):
        args, chunk = data
        args = object_from_shared_memory(args)
        logging.info("Processing chunk %s" % str(chunk))

        t = time.perf_counter()
        kmer_finder = DenseKmerFinder(args["graph"], args["kmer_size"], critical_graph_paths=args["critical_graph_paths"],
                                      position_id=args["position_id"],
                                      max_variant_nodes=args["max_variant_nodes"],
                                      only_save_one_node_per_kmer=True,
                                      start_at_critical_path_number=chunk[0],
                                      stop_at_critical_path_number=chunk[1],
                                      whitelist=args["whitelist"])
        kmer_finder.find()
        logging.info("Took %d sec to find kmers" % (time.perf_counter()-t))
        res = kmer_finder.get_flat_kmers(v="1")
        if len(np.where(res._hashes == 979349781162838629)[0]) > 0:
            logging.info("FOUND MATCH! %d/%d" % (chunk[0], chunk[1]))

        return res

    def index(args):
        pool = get_shared_pool(args.n_threads)

        if args.critical_graph_paths is None:
            logging.info("Making critical graph paths since not specified")
            args.critical_paths = CriticalGraphPaths.from_graph(graph, args.k)

        args.position_id = PositionId.from_graph(args.graph) if args.position_id is not None else None

        args = vars(args)
        args.pop("func")
        args_shared = object_to_shared_memory(args)
        critical_paths = args["critical_graph_paths"]

        n_chunks = args["n_threads"]*20
        if n_chunks >= len(critical_paths):
            n_chunks = len(critical_paths)

        assert n_chunks <= len(critical_paths), "Too many threads for data. Try with fewer"

        logging.info("There are %d critical paths" % len(critical_paths))
        logging.info("Will process these in %d chunks" % n_chunks)
        n_paths_per_chunk = len(critical_paths) // n_chunks
        logging.info("N paths per chunk: %d" % n_paths_per_chunk)

        starting_indexes = list(range(0, len(critical_paths), n_paths_per_chunk))
        ending_indexes = starting_indexes[1:] + [len(critical_paths)]
        chunks = list((s, e) for s, e in zip(starting_indexes, ending_indexes))

        logging.info("Will process chunks (from,to): %s" % chunks)

        t = time.perf_counter()
        flat_kmers = []
        for i, results in enumerate(pool.imap(index_single_thread, zip(itertools.repeat(args_shared), chunks))):
            flat_kmers.append(results)
            logging.info("Done with chunk. Found %d kmers" % (len(results._hashes)))

        logging.info("Time spent to make indexes: %.2f" % (time.perf_counter()-t))
        close_shared_pool()

        flat_kmers = FlatKmers.from_multiple_flat_kmers(flat_kmers)
        logging.info("N kmers in flat kmers: %d" % len(flat_kmers._hashes))
        if args["include_reverse_complement"]:
            logging.info("Adding reverse complements")
            revcomp_flat_kmers = flat_kmers.get_reverse_complement_flat_kmers(args["kmer_size"])
            flat_kmers = FlatKmers.from_multiple_flat_kmers([flat_kmers, revcomp_flat_kmers])
            logging.info("N kmers in flat kmers after revcomp added: %d" % len(flat_kmers._hashes))

        flat_kmers.to_file(args["out_file_name"])
        logging.info("Wrote final FlatKmers to %s" % args["out_file_name"])


    subparser = subparsers.add_parser("index")
    subparser.add_argument("-g", "--graph", required=True, type=Graph.from_file)
    subparser.add_argument("-c", "--critical_graph_paths", required=False, type=from_file, help="Will be created if not specified")
    subparser.add_argument("-p", "--position_id", required=False, type=from_file, help="Will be created if not specified")
    subparser.add_argument("-k", "--kmer-size", type=int, default=31, required=False)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-t", "--n-threads", required=False, type=int, default=1)
    #subparser.add_argument("-w", "--whitelist", required=False, type=lambda e: NpSet(FlatKmers.from_file(e)._hashes), help="Only store kmers that are in the whitelist")
    subparser.add_argument("-w", "--whitelist", required=False, type=lambda e: CollisionFreeKmerIndex.from_file(e), help="Only store kmers that are in the whitelist")
    subparser.add_argument("-r", "--include-reverse-complement", required=False, type=bool, default=False)
    subparser.add_argument("-O", "--only-save-one-node-per-kmer", required=False, type=bool, default=False)
    subparser.add_argument("-v", "--max-variant-nodes", required=False, type=int, default=5, help="Max variant nodes allowed in kmer.")
    subparser.set_defaults(func=index)


    def find_critical_paths(args):
        from .critical_graph_paths import CriticalGraphPaths
        graph = Graph.from_file(args.graph)
        critical_paths = CriticalGraphPaths.from_graph(graph, args.kmer_size)
        critical_paths._make_index()
        to_file(critical_paths, args.out_file_name)
        logging.info("Wrote to file %s" % args.out_file_name)

    subparser = subparsers.add_parser("find_critical_paths")
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-k", "--kmer-size", type=int, default=31, required=False)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=find_critical_paths)


    def add_reverse_complements(args):
        flat = FlatKmers.from_file(args.flat_kmers)
        rev_comp_flat = flat.get_reverse_complement_flat_kmers(k=args.kmer_size)
        flat = FlatKmers.from_multiple_flat_kmers([flat, rev_comp_flat])
        flat.to_file(args.out_file_name)
        logging.info("Saved new flat kmers with reverse complements to %s" % args.out_file_name)

    subparser = subparsers.add_parser("add_reverse_complements")
    subparser.add_argument("-f", "--flat-kmers", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-k", "--kmer-size", type=int, required=True)
    subparser.set_defaults(func=add_reverse_complements)


    def count_kmers(args):
        from .kmer_counter import KmerCounter
        counter = KmerCounter.from_flat_kmersv2(args.flat_kmers, args.modulo, args.subsample_ratio)
        to_file(counter, args.out_file_name)
        logging.info("Wrote counter to %s" % args.out_file_name)

    subparser = subparsers.add_parser("count_kmers")
    subparser.add_argument("-f", "--flat-kmers", required=True, type=FlatKmers.from_file)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-m", "--modulo", required=False, type=int, default=200000033)
    subparser.add_argument("-s", "--subsample-ratio", required=False, type=int, default=1, help="1 to keep every kmer, 2 for every other etc")
    subparser.set_defaults(func=count_kmers)


    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    args.func(args)
    remove_shared_memory_in_session()




import pickle
import logging
import numpy as np
import itertools
from collections import defaultdict


class UniqueKmerIndex:
    def __init__(self, index_dict):
        self._index_dict = index_dict

    def to_file(self, file_name):
        file = open(file_name, "wb")
        pickle.dump(self._index_dict, file)

    @classmethod
    def from_file(cls, file_name):
        with open(file_name, "rb") as f:
            data = pickle.load(f)

        return cls(data)

    def get(self, kmer):
        return self._index_dict[kmer]

    @classmethod
    def from_flat_kmers_and_snps_graph(cls, flat_kmers, graph, reverse_index):
        g = graph
        flat = flat_kmers

        logging.info("Finding kmers in flat index that are unique in regards to starting position and kmers")
        no_duplicates = []
        found = set()

        for i in range(len(flat._hashes)):
            hash = flat._hashes[i]
            ref_pos = flat._ref_offsets[i]

            if (hash, ref_pos) not in found:
                no_duplicates.append(hash)
                found.add((hash, ref_pos))

            if i % 100000 == 0:
                print("%d of %d kmers procsesed" % (i, len(flat._hashes)))

        logging.info("Number of no kmers that are not duplicate in index: %d" % len(no_duplicates))
        no_duplicates = np.array(no_duplicates)

        logging.info("Finding snps in graph")
        snps = [g.adj_list[node] for node in g.blocks if len(g.adj_list[node]) == 2]
        logging.info("Finding kmers that occur only once in index")
        unique_no_duplicates, positions = np.unique(no_duplicates, return_counts=True)
        unique_kmers = set(unique_no_duplicates[np.where(positions == 1)[0]])

        logging.info("Creating index as SNPs where both nodes only have kmers thare are unique in index")

        has_unique = [np.all([kmer in unique_kmers for kmer in
                              itertools.chain(reverse_index.get_node_kmers(nodes[0]), reverse_index.get_node_kmers(nodes[1]))])
                      if len(reverse_index.get_node_kmers(nodes[0])) < 4 and len(reverse_index.get_node_kmers(nodes[1])) < 4 else False
                      for
                      nodes in snps]

        snps_array = np.array(snps)
        snps_with_unique_kmers = snps_array[np.array(has_unique)]

        unique_index = defaultdict(list)
        for i, node_pair in enumerate(snps_with_unique_kmers):
            for node in node_pair:
                for kmer in reverse_index.get_node_kmers(node):
                    unique_index[kmer].append(node)

            if i % 1000 == 0:
                logging.info("Creating index. %d/%d variants processed " % (len(unique_index), len(snps_with_unique_kmers)))


        return cls(unique_index)



import logging
from .bionumpy_wrapper import bionumpy_hash
from .flat_kmers import FlatKmers
import numpy as np

def sample_kmers_from_structural_variants(graph, variant_to_nodes, kmer_index_with_frequencies, k, max_frequency=2):
    """"
    For every variant with big nodes,
    try to find kmers within the node with low frequency in kmer_index_with_frequencies
    """

    kmers_found = []
    nodes_found = []
    ref_offsets_found = []

    for ref_node, var_node  in variant_to_nodes:
        for node in (ref_node, var_node):
            if graph.get_node_size(node) > k + 5:
                node_sequence = graph.get_numeric_node_sequence(node)
                node_kmers = bionumpy_hash(node_sequence, k)
                kmer_frequencies = np.array([kmer_index_with_frequencies.get_frequency(k) for k in node_kmers])
                valid = np.where(kmer_frequencies < max_frequency)[0]

                # do not choose overlapping kmers
                chosen = []
                prev = -10000
                for v in valid:
                    if v >= prev + k:
                        chosen.append(v)
                        prev = v

                #logging.info("Node kmers: %s" % node_kmers)
                #logging.info("Chose positions %s with kmers %s for node %d" % (chosen, node_kmers[chosen], node))

                if len(chosen) > 0:
                    kmers_found.extend(node_kmers[chosen])
                    nodes_found.extend([node]*len(chosen))
                    ref_offsets_found.extend([0]*len(chosen))


    return FlatKmers(np.array(kmers_found, dtype=np.uint64),
                     np.array(nodes_found, dtype=np.uint32),
                     np.array(ref_offsets_found, dtype=np.uint32))


import numpy as np

class VariantNodesIndex:
    def __init__(self, ref_positions, variant_nodes):
        self.ref_position = ref_positions
        self.variant_nodes = variant_nodes

    @classmethod
    def from_file(cls, file_name):
        data = np.load(file_name)
        return cls(data["ref_positions"], data["variant_nodes"])

    def get_variant_nodes_between_ref_positions(self, ref_start, ref_end):


    @classmethod
    def from_graph(cls, graph):
        pass

    def to_file(self, file_name):
        np.savez(file_name, ref_position=self.ref_position, variant_nodes=self.variant_nodes)



import numpy as np
import logging


class CriticalGraphPaths:
    def __init__(self, nodes, offsets, index=None):
        self.nodes = nodes
        self.offsets = offsets
        self._index = index

    def _make_index(self):
        if len(self.nodes) == 0:
            self._index = np.zeros(0)
            return

        self._index = np.zeros(np.max(self.nodes)+1, dtype=np.uint16)
        self._index[self.nodes] = self.offsets

    @classmethod
    def empty(cls):
        return cls(np.array([]), np.array([]), np.array([]))

    def is_critical(self, node, offset):
        if self._index is None:
            logging.info("Making critical paths index")
            logging.info("NOdes: %s, offsets: %s" % (self.nodes, self.offsets))
            self._make_index()

        if node >= len(self._index):
            return False

        return self._index[node] == offset

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        return ((node, offset) for node, offset in zip(self.nodes, self.offsets))

    @classmethod
    def from_graph(cls, graph, k):

        logging.info("Getting reverse edges")
        reverse_edges = graph.get_reverse_edges_hashtable()

        critical_nodes = []
        critical_offsets = []


        logging.info("Iterating chromosome start nodes %s" % graph.chromosome_start_nodes)
        for start_node in graph.chromosome_start_nodes:

            #logging.info("On chromosome start node %d" % start_node)
            current_node = start_node
            depth = 0
            bp_since_last_join = 0
            i = 0
            while True:
                if i % 100000 == 0:
                    logging.info("%d / %d nodes traversed. %d critical paths found." % (i, graph.max_node_id(), len(critical_nodes)))
                i += 1

                # lower depth with number of edges coming in
                prev_depth = depth
                depth -= len(reverse_edges[current_node])
                if prev_depth > 1 and depth == 0:
                    #print("Resetting bp since last join, Prev depth: %d" % prev_depth)
                    bp_since_last_join = 0

                #print("Node %d, depth %d. Bp since last join: %d" % (current_node, depth, bp_since_last_join))

                node_size = graph.get_node_size(current_node)

                if depth == 0 and node_size != 0:  # ignore dummy nodes
                    #if bp_since_last_join + node_size >= k:
                    if bp_since_last_join <= k and bp_since_last_join + node_size >= k:
                        # we have a critical path point somewhere on this node
                        critical_nodes.append(current_node)
                        #print("  adding node %d offset %d. Bp since last join is now %d" % (current_node, k-bp_since_last_join-1, bp_since_last_join))
                        critical_offsets.append(k - bp_since_last_join-1)
                        #bp_since_last_join += node_size

                next_nodes = graph.get_edges(current_node)
                depth += len(next_nodes)

                if len(next_nodes) == 0:
                    break
                elif len(next_nodes) == 1:
                    bp_since_last_join += node_size
                    #print(" single next node, increased bp since last join to %d (node size of %d is %d) " % (bp_since_last_join, current_node, node_size))
                    current_node = next_nodes[0]
                else:
                    next_nodes = [n for n in next_nodes if graph.is_linear_ref_node_or_linear_ref_dummy_node(n)]
                    if len(next_nodes) != 1:
                        logging.error("Did not find 1 next node from node %d" % current_node)
                        logging.error("Edges: %s" % graph.get_edges(current_node))
                        raise Exception("")
                    current_node = next_nodes[0]

        logging.info("Found %d critical paths" % len(critical_nodes))
        return cls(np.array(critical_nodes, dtype=np.uint32), np.array(critical_offsets, np.uint16))


